from __future__ import division
from __future__ import print_function

from os import makedirs, path
import numpy as np

def default_metric(curriculum_params, input_data, correct_output, output_mask, output, epoch, losses, verbosity):
	""" Default metric to use to evaluate performance when using Curriculum learning.

	Advance is true if accuracy >= threshold, False otherwise.

	Arguments:
		curriculum_params (dict): Dictionary of the :class:`Curriculum` object parameters, containing the following keys:

			:Dictionary Keys:
				* **stop_training** (*bool*) -- True if the network has finished training and completed all stages.
				* **stage** (*int*) -- Current training stage (initial stage is 0).
				* **metric_values** (*list of [float, int]*)  -- List of metric values and the stage at which each metric value was computed.
				* **tasks** (*list of :class:`psychrnn.tasks.task.Task` objects*) -- List of tasks in the curriculum.
				* **metric** (*function*) -- What metric function to use. :func:`default_metric` is an example of one in terms of inputs and outputs taken.
				* **accuracies** (*list of functions with the signature of* :func:`psychrnn.tasks.task.Task.accuracy_function`) -- Accuracy function to use at each stage.
				* **thresholds** (*list of float*) -- Thresholds for each stage that accuracy must reach to move to the next stage.
				* **metric_epoch** (*int*) -- Calculate the metric / test if advance to the next stage every metric_epoch training epochs.
				* **output_file** (*str*) -- Optional path for where to save out metric value and stage.

		input_data (ndarray(dtype=float, shape =(:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` ))): Task inputs.
		correct_output (ndarray(dtype=float, shape = (:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out`))): Correct (target) task output given input_data.
		output_mask (ndarray(dtype=float, shape = (:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out`))): Output mask for the task. True when the network should aim to match the target output, False when the target output can be ignored.
		output (ndarray(dtype=float, shape = (:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out`))): The network's output given input_data.
		epoch (*int*): The epoch number in training.
		losses (*list of float*): List of losses, computed during training.
		verbosity (*bool*): Whether to print information as training progresses. If True, prints accuracy every time it is computed.

	Returns:
		tuple:

		* **advance** (*bool*) -- True if the accuracy is >= the threshold for the current stage. False otherwise.
		* **metric_value** (*float*) -- Value of the computed accuracy.

	"""
	accuracy = curriculum_params['accuracies'][curriculum_params['stage']](correct_output,output, output_mask)
	threshold = curriculum_params['thresholds'][curriculum_params['stage']]
	if verbosity:
		print("Accuracy: " + str(accuracy))
	return accuracy>=threshold, accuracy

class Curriculum(object):
	""" Curriculum object.

	Allows training on a sequence of tasks when Curriculum is passed into :func:`~psychrnn.backend.rnn.RNN.train`.

	Arguments:
		tasks (list of :class:`~psychrnn.tasks.task.Task` objects):  List of tasks  to use in the curriculum.
		metric (function, optional): Function for calculating whether the stage advances and what the metric value is at each metric_epoch. Default: :func:`default_metric`.

			:Arguments:
				* **curriculum_params** (*dict*) -- Dictionary of the :class:`Curriculum` object parameters, containing the following keys:

				:Dictionary Keys:
					* **stop_training** (*bool*) -- True if the network has finished training and completed all stages.
					* **stage** (*int*) -- Current training stage (initial stage is 0).
					* **metric_values** (*list of [float, int]*)  -- List of metric values and the stage at which each metric value was computed.
					* **tasks** (*list of :class:`psychrnn.tasks.task.Task` objects*) -- List of tasks in the curriculum.
					* **metric** (*function*) -- What metric function to use. :func:`default_metric` is an example of one in terms of inputs and outputs taken.
					* **accuracies** (*list of functions*) -- Accuracy function to use at each stage.
					* **thresholds** (*list of float*) -- Thresholds for each stage that accuracy must reach to move to the next stage.
					* **metric_epoch** (*int*) -- Calculate the metric and test if the model should advance to the next stage every :data:`metric_epoch` training epochs.
					* **output_file** (*str*) -- Optional path for saving out themetric value and stage. If the .npz filename extension is not included, it will be appended.

				* **input_data** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Task inputs.
				* **correct_output** (*ndarray(dtype=float, shape = (*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Correct (target) task output given input_data.
				* **output_mask** (*ndarray(dtype=float, shape = (*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Output mask for the task. True when the network should aim to match the target output, False when the target output can be ignored.
				* **output** (*ndarray(dtype=float, shape = (*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- The network's output given input_data.
				* **epoch** (*int*) -- The epoch number in training.
				* **losses** (*list of float*) -- List of losses, computed during training.
				* **verbosity** (*bool*) -- Whether to print information as training progresses.

			:Returns:
				*tuple*

				* **advance** (*bool*) -- True if the the stage should be advanced. False otherwise.
				* **metric_value** (*float*) -- Value of the computed metric.

		accuracies (list of functions, optional): Optional list of functions to use to calculate network performance for the purposes of advancing tasks. Used by :func:`default_metric` to compute accuracy. Default: ``[tasks[i].accuracy_function for i in range(len(tasks))]``.
		thresholds (list of float, optional): Optional list of thresholds. If metric = default_metric, accuracies must reach the threshold for a given stage in order to advance to the next stage. Default: ``[.9 for i in range(len(tasks))]``
		metric_epoch (int): Calculate the metric and test if the model should advance to the next stage every :data:`metric_epoch` training epochs. Default: 10
		output_file (str): Optional path for saving out the metric value and stage. If the .npz filename extension is not included, it will be appended. Default: None.

	"""
	def __init__(self, tasks, **kwargs):
		self.stop_training = False
		self.stage = 0
		self.metric_values = []

		# List of tasks that make up the curriculum
		self.tasks = tasks

		#Optional function with parameters as in default_metric that returns whether to advance stage, and the accuracy / metric value
		self.metric = kwargs.get('metric', default_metric)
		
		#Optional list of accuracy functions to use for each task
		self.accuracies = kwargs.get('accuracies', [tasks[i].accuracy_function for i in range(len(tasks))])
		assert len(self.accuracies)==len(self.tasks)
		
		# Optional list of accuracy cuttoff values to use with each tasks
		self.thresholds = kwargs.get('thresholds', [.9 for i in range(len(tasks))])
		assert len(self.thresholds)==len(self.tasks)
		
		# How often to check metric?
		self.metric_epoch = kwargs.get('metric_epoch', 10)
		
		# Optional path to save out metric value and stage to
		self.output_file = kwargs.get('output_file', None)
		if self.output_file is not None:
			if path.dirname(self.output_file) != "" and not path.exists(path.dirname(self.output_file)):
				makedirs(path.dirname(self.output_file))
	
	def metric_test(self, input_data, correct_output, output_mask, test_output, epoch, losses, verbosity = False):
		"""Evaluates whether to advance the stage to the next task or not.

		Arguments:
			input_data (ndarray(dtype=float, shape =(:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` ))): Task inputs.
			correct_output (ndarray(dtype=float, shape = (:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out`))): Correct (target) task output given input_data.
			output_mask (ndarray(dtype=float, shape = (:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out`))): Output mask for the task. True when the network should aim to match the target output, False when the target output can be ignored.
			test_output (ndarray(dtype=float, shape = (:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out`))): The network's output given input_data.
			epoch (*int*): The epoch number in training.
			losses (*list of float*): List of losses, computed during training.
			verbosity (*bool, optional*): Whether to print information as metric is computed and stages advanced. Default: False

		Returns:
			True if stage advances, False otherwise.
		"""
		advance, metric_value = self.metric(self.__dict__, input_data, correct_output, output_mask, test_output, epoch, losses, verbosity)
		self.metric_values.append([metric_value, self.stage])
		if advance:
			self.stage+=1
			if self.stage == len(self.tasks):
				self.stop_training = True
				if self.output_file is not None:
					np.save(self.output_file, self.metric_values)
					if verbosity:
						print("Metric values saved in file: %s" % self.output_file)
			if verbosity:
				print("Stage " + str(self.stage))
			return True
		return False

	def get_generator_function(self):
		""" Return the generator function for the current task.

		Returns:
			:func:`psychrnn.tasks.task.Task.batch_generator` function: Task batch generator for the task at the current stage.
		"""
		return self.tasks[self.stage].batch_generator()




from __future__ import division

import numpy as np

from abc import ABCMeta, abstractmethod

# abstract class python 2 & 3 compatible
ABC = ABCMeta('ABC', (object,), {})


class Task(ABC):
    """ The base task class.

    The base task class provides the structure that users can use to\
    define a new task. This structure is used by example tasks \
    :class:`~psychrnn.tasks.perceptual_discrimination.PerceptualDiscrimination`, \
    :class:`~psychrnn.tasks.match_to_category.MatchToCategory`, \
    and :class:`~psychrnn.tasks.delayed_discrim.DelayedDiscrimination`.

    Note:
        The base task class is not itself a functioning task. 
        The generate_trial_params and trial_function must be defined to define a new, functioning, task.

    Args:
        N_in (int): The number of network inputs.
        N_out (int): The number of network outputs.
        dt (float): The simulation timestep.
        tau (float): The intrinsic time constant of neural state decay.
        T (float): The trial length.
        N_batch (int): The number of trials per training update.

    Inferred Parameters:
        * **alpha** (*float*) -- The number of unit time constants per simulation timestep.
        * **N_steps** (*int*): The number of simulation timesteps in a trial. 

    """
    def __init__(self, N_in, N_out, dt, tau, T, N_batch):

        # ----------------------------------
        # Initialize required parameters
        # ----------------------------------
        self.N_batch = N_batch
        self.N_in = N_in
        self.N_out = N_out
        self.dt = dt
        self.tau = tau
        self.T = T

        # ----------------------------------
        # Calculate implied parameters
        # ----------------------------------
        self.alpha = (1.0 * self.dt) / self.tau
        self.N_steps = int(np.ceil(self.T / self.dt))

    def get_task_params(self):
        """ Get dictionary of task parameters.

        Note:
            N_in, N_out, N_batch, dt, tau and N_steps must all be passed to the network model as parameters -- this function is the recommended way to begin building the network_params that will be passed into the RNN model.


        Returns:
            dict: Dictionary of :class:`Task` attributes including the following keys:

            :Dictionary Keys: 
                * **N_batch** (*int*) -- The number of trials per training update.
                * **N_in** (*int*) -- The number of network inputs.
                * **N_out** (*int*) -- The number of network outputs.
                * **dt** (*float*) -- The simulation timestep.
                * **tau** (*float*) -- The unit time constant.
                * **T** (*float*) -- The trial length.
                * **alpha** (*float*) -- The number of unit time constants per simulation timestep.
                * **N_steps** (*int*): The number of simulation timesteps in a trial. 

            Note:
                The dictionary will also include any other attributes defined in your task definition.
        
        """
        return self.__dict__
    
    @abstractmethod
    def generate_trial_params(self, batch, trial):
        """ Define parameters for each trial.

        Using a combination of randomness, presets, and task attributes, define the necessary trial parameters.

        Args:
            batch (int): The batch number for this trial.
            trial (int): The trial number of the trial within the batch data:`batch`.

        Returns:
            dict: Dictionary of trial parameters.


        Warning:
            This function is abstract and must be implemented in a child Task object.

        Example:
            See :func:`PerceptualDiscrimination <psychrnn.tasks.perceptual_discrimination.PerceptualDiscrimination.generate_trial_params>`,\
            :func:`MatchToCategory <psychrnn.tasks.match_to_category.MatchToCategory.generate_trial_params>`,\
            and :func:`DelayedDiscrimination <psychrnn.tasks.delayed_discrim.DelayedDiscrimination.generate_trial_params>` for example implementations.

        """
        pass

    @abstractmethod
    def trial_function(self, time, params):
        """ Compute the trial properties at :data:`time`.

        Based on the :data:'params' compute the trial stimulus (x_t), correct output (y_t), and mask (mask_t) at :data:`time`.

        Args:
            time (int): The time within the trial (0 <= :data:`time` < :attr:`T`).
            params (dict): The trial params produced by :func:`~psychrnn.tasks.task.Task.generate_trial_params`

        Returns:
            tuple:

            * **x_t** (*ndarray(dtype=float, shape=(*:attr:`N_in` *,))*) -- Trial input at :data:`time` given :data:`params`.
            * **y_t** (*ndarray(dtype=float, shape=(*:attr:`N_out` *,))*) -- Correct trial output at :data:`time` given :data:`params`.
            * **mask_t** (*ndarray(dtype=bool, shape=(*:attr:`N_out` *,))*) -- True if the network should train to match the y_t, False if the network should ignore y_t when training.

        Warning:
            This function is abstract and must be implemented in a child Task object.

        Example:
            See :func:`PerceptualDiscrimination <psychrnn.tasks.perceptual_discrimination.PerceptualDiscrimination.trial_function>`,\
            :func:`MatchToCategory <psychrnn.tasks.match_to_category.MatchToCategory.trial_function>`,\
            and :func:`DelayedDiscrimination <psychrnn.tasks.delayed_discrim.DelayedDiscrimination.trial_function>` for example implementations.
        
        """
        pass

    
    def accuracy_function(self, correct_output, test_output, output_mask):
        """ Function to calculate accuracy (not loss) as it would be measured experimentally.

        Output should range from 0 to 1. This function is used by :class:`~psychrnn.backend.curriculum.Curriculum` as part of it's :func:`~psychrnn.backend.curriculum.default_metric`.

        Args:
            correct_output(ndarray(dtype=float, shape =(:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` ))): Correct batch output. ``y_data`` as returned by :func:`batch_generator`.
            test_output(ndarray(dtype=float, shape =(:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` ))): Output to compute the accuracy of. ``output`` as returned by :func:`psychrnn.backend.rnn.RNN.test`.
            output_mask(ndarray(dtype=bool, shape =(:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out`))): Mask. ``mask`` as returned by func:`batch_generator`.

        Returns:
            float: 0 <= accuracy <=1

        Warning:
            This function is abstract and may optionally be implemented in a child Task object.

        Example:
            See :func:`PerceptualDiscrimination <psychrnn.tasks.perceptual_discrimination.PerceptualDiscrimination.accuracy_function>`,\
            :func:`MatchToCategory <psychrnn.tasks.match_to_category.MatchToCategory.accuracy_function>`,\
            and :func:`DelayedDiscrimination <psychrnn.tasks.delayed_discrim.DelayedDiscrimination.accuracy_function>` for example implementations.
        """
        pass

    def generate_trial(self, params):
        """ Loop to generate a single trial.

        Args:
            params(dict): Dictionary of trial parameters generated by :func:`generate_trial_params`.

        Returns:
            tuple:

            * **x_trial** (*ndarray(dtype=float, shape=(*:attr:`N_steps`, :attr:`N_in` *))*) -- Trial input given :data:`params`.
            * **y_trial** (*ndarray(dtype=float, shape=(*:attr:`N_steps`, :attr:`N_out` *))*) -- Correct trial output given :data:`params`.
            * **mask_trial** (*ndarray(dtype=bool, shape=(*:attr:`N_steps`, :attr:`N_out` *))*) -- True during steps where the network should train to match :data:`y`, False where the network should ignore :data:`y` during training.
        """

        # ----------------------------------
        # Loop to generate a single trial
        # ----------------------------------
        x_data = np.zeros([self.N_steps, self.N_in])
        y_data = np.zeros([self.N_steps, self.N_out])
        mask = np.zeros([self.N_steps, self.N_out])

        for t in range(self.N_steps):
            x_data[t, :], y_data[t, :], mask[t, :] = self.trial_function(t * self.dt, params)

        return x_data, y_data, mask

    def batch_generator(self):
        """ Generates a batch of trials.

        Returns:
            Generator[tuple, None, None]:

        Yields:
            tuple:

            * **stimulus** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Task stimuli for :attr:`N_batch` trials.
            * **target_output** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Target output for the network on :attr:`N_batch` trials given the :data:`stimulus`.
            * **output_mask** (*ndarray(dtype=bool, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Output mask for :attr:`N_batch` trials. True when the network should aim to match the target output, False when the target output can be ignored.
            * **trial_params** (*ndarray(dtype=dict, shape =(*:attr:`N_batch` *,))*): Array of dictionaries containing the trial parameters produced by :func:`generate_trial_params` for each trial in :attr:`N_batch`.
        
        """

        batch = 1
        while batch > 0:

            x_data = []
            y_data = []
            mask = []
            params = []
            # ----------------------------------
            # Loop over trials in batch
            # ----------------------------------
            for trial in range(self.N_batch):
                # ---------------------------------------
                # Generate each trial based on its params
                # ---------------------------------------
                p = self.generate_trial_params(batch, trial)
                x,y,m = self.generate_trial(p)
                x_data.append(x)
                y_data.append(y)
                mask.append(m)
                params.append(p)

            batch += 1

            yield np.array(x_data), np.array(y_data), np.array(mask), np.array(params)

    def get_trial_batch(self):
        """Get a batch of trials.

        Wrapper for :code:`next(self.batch_generator())`.

        Returns:
            tuple:

            * **stimulus** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_in` *))*): Task stimuli for :attr:`N_batch` trials.
            * **target_output** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Target output for the network on :attr:`N_batch` trials given the :data:`stimulus`.
            * **output_mask** (*ndarray(dtype=bool, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Output mask for :attr:`N_batch` trials. True when the network should aim to match the target output, False when the target output can be ignored.
            * **trial_params** (*ndarray(dtype=dict, shape =(*:attr:`N_batch` *,))*): Array of dictionaries containing the trial parameters produced by :func:`generate_trial_params` for each trial in :attr:`N_batch`.

        """
        return next(self.batch_generator())




from __future__ import division

from psychrnn.tasks.task import Task
import numpy as np

class PerceptualDiscrimination(Task):
    """Two alternative forced choice (2AFC) binary discrimination task. 

    On each trial the network receives two simultaneous noisy inputs into each of two input channels. The network must determine which channel has the higher mean input and respond by driving the corresponding output unit to 1.

    Takes two channels of noisy input (:attr:`N_in` = 2).
    Two channel output (:attr:`N_out` = 2) with a one hot encoding (high value is 1, low value is .2) towards the higher mean channel.

    Loosely based on `Britten, Kenneth H., et al. "The analysis of visual motion: a comparison of neuronal and psychophysical performance." Journal of Neuroscience 12.12 (1992): 4745-4765 <https://www.jneurosci.org/content/12/12/4745>`_

    Args:
        dt (float): The simulation timestep.
        tau (float): The intrinsic time constant of neural state decay.
        T (float): The trial length.
        N_batch (int): The number of trials per training update.
        coherence (float, optional): Amount by which the means of the two channels will differ. By default None.
        direction (int, optional): Either 0 or 1, indicates which input channel will have higher mean input. By default None.

    """

    def __init__(self, dt, tau, T, N_batch, coherence = None, direction = None):
        super(PerceptualDiscrimination,self).__init__(2, 2, dt, tau, T, N_batch)
        
        self.coherence = coherence

        self.direction = direction

        self.lo = 0.2 # Low value for one hot encoding

        self.hi = 1.0 # High value for one hot encoding

    def generate_trial_params(self, batch, trial):
        """Define parameters for each trial.

        Implements :func:`~psychrnn.tasks.task.Task.generate_trial_params`.

        Args:
            batch (int): The batch number that this trial is part of.
            trial (int): The trial number of the trial within the batch *batch*.

        Returns:
            dict: Dictionary of trial parameters including the following keys:

            :Dictionary Keys: 
                * **coherence** (*float*) -- Amount by which the means of the two channels will differ. :attr:`self.coherence` if not None, otherwise ``np.random.exponential(scale=1/5)``.
                * **direction** (*int*) -- Either 0 or 1, indicates which input channel will have higher mean input. :attr:`self.direction` if not None, otherwise ``np.random.choice([0, 1])``.
                * **stim_noise** (*float*) -- Scales the stimlus noise. Set to .1.
                * **onset_time** (*float*) -- Stimulus onset time. ``np.random.random() * self.T / 2.0``.
                * **stim_duration** (*float*) -- Stimulus duration. ``np.random.random() * self.T / 4.0 + self.T / 8.0``.

        """

        # ----------------------------------
        # Define parameters of a trial
        # ----------------------------------
        params = dict()
        if self.coherence == None:
            params['coherence'] = np.random.choice([0.1, 0.3, 0.5, 0.7])
        else:
            params['coherence'] = self.coherence
        params['direction'] = np.random.choice([0, 1])
        params['stim_noise'] = 0.1
        params['onset_time'] = np.random.random() * self.T / 2.0
        params['stim_duration'] = np.random.random() * self.T / 4.0 + self.T / 8.0

        return params

    def trial_function(self, t, params):
        """Compute the trial properties at :data:`time`.

        Implements :func:`~psychrnn.tasks.task.Task.trial_function`.

        Based on the :data:`params` compute the trial stimulus (x_t), correct output (y_t), and mask (mask_t) at :data:`time`.

        Args:
            time (int): The time within the trial (0 <= :data:`time` < :attr:`T`).
            params (dict): The trial params produced by :func:`generate_trial_params`.

        Returns:
            tuple:

            * **x_t** (*ndarray(dtype=float, shape=(*:attr:`N_in` *,))*) -- Trial input at :data:`time` given :data:`params`. For ``params['onset_time'] < time < params['onset_time'] + params['stim_duration']`` , 1 is added to the noise in both channels, and :data:`params['coherence']` is also added in the channel corresponding to :data:`params[dir]`.
            * **y_t** (*ndarray(dtype=float, shape=(*:attr:`N_out` *,))*) -- Correct trial output at :data:`time` given :data:`params`. From ``time > params['onset_time'] + params[stim_duration] + 20`` onwards, the correct output is encoded using one-hot encoding. Until then, y_t is 0 in both channels.
            * **mask_t** (*ndarray(dtype=bool, shape=(*:attr:`N_out` *,))*) -- True if the network should train to match the y_t, False if the network should ignore y_t when training. The mask is True for ``time > params['onset_time'] + params['stim_duration']`` and False otherwise.

        """

        # ----------------------------------
        # Initialize with noise
        # ----------------------------------
        x_t = np.sqrt(2*.01*np.sqrt(10)*np.sqrt(self.dt)*params['stim_noise']*params['stim_noise'])*np.random.randn(self.N_in)
        y_t = np.zeros(self.N_out)
        mask_t = np.ones(self.N_out)

        # ----------------------------------
        # Retrieve parameters
        # ----------------------------------
        coh = params['coherence']
        onset = params['onset_time']
        stim_dur = params['stim_duration']
        dir = params['direction']

        # ----------------------------------
        # Compute values
        # ----------------------------------
        if onset < t < onset + stim_dur:
            x_t[dir] += 1 + coh
            x_t[(dir + 1) % 2] += 1

        if t > onset + stim_dur + 20:
            y_t[dir] = self.hi
            y_t[1-dir] = self.lo

        if t < onset + stim_dur:
            mask_t = np.zeros(self.N_out)

        return x_t, y_t, mask_t

    def accuracy_function(self, correct_output, test_output, output_mask):
        """Calculates the accuracy of :data:`test_output`.

        Implements :func:`~psychrnn.tasks.task.Task.accuracy_function`.

        Takes the channel-wise mean of the masked output for each trial. Whichever channel has a greater mean is considered to be the network's "choice".

        Returns:
            float: 0 <= accuracy <= 1. Accuracy is equal to the ratio of trials in which the network made the correct choice as defined above.
        
        """


        chosen = np.argmax(np.mean(test_output*output_mask, axis=1), axis = 1)
        truth = np.argmax(np.mean(correct_output*output_mask, axis = 1), axis = 1)
        return np.mean(np.equal(truth, chosen))




from __future__ import division
from __future__ import print_function

from abc import ABCMeta, abstractmethod

# abstract class python 2 & 3 compatible
ABC = ABCMeta('ABC', (object,), {})

import tensorflow as tf
import numpy as np

import sys
from time import time
from os import makedirs, path
from inspect import isgenerator

from psychrnn.backend.regularizations import Regularizer
from psychrnn.backend.loss_functions import LossFunction
from psychrnn.backend.initializations import WeightInitializer, GaussianSpectralRadius

tf.compat.v1.disable_eager_execution()

class RNN(ABC):
    """ The base recurrent neural network class.

    Note:
        The base RNN class is not itself a functioning RNN. 
        forward_pass must be implemented to define a functioning RNN.

    Args:
       params (dict): The RNN parameters. Use your tasks's :func:`~psychrnn.tasks.task.Task.get_task_params` function to start building this dictionary. Optionally use a different network's :func:`get_weights` function to initialize the network with preexisting weights.

       :Dictionary Keys: 
            * **name** (*str*) -- Unique name used to determine variable scope. Having different variable scopes allows multiple distinct models to be instantiated in the same TensorFlow environment. See `TensorFlow's variable_scope <https://www.tensorflow.org/api_docs/python/tf/compat/v1/variable_scope>`_ for more details.
            * **N_in** (*int*) -- The number of network inputs.
            * **N_rec** (*int*) -- The number of recurrent units in the network.
            * **N_out** (*int*) -- The number of network outputs.
            * **N_steps** (*int*): The number of simulation timesteps in a trial. 
            * **dt** (*float*) -- The simulation timestep.
            * **tau** (*float*) -- The intrinsic time constant of neural state decay.
            * **N_batch** (*int*) -- The number of trials per training update.
            * **rec_noise** (*float, optional*) -- How much recurrent noise to add each time the new state of the network is calculated. Default: 0.0.
            * **transfer_function** (*function, optional*) -- Transfer function to use for the network. Default: `tf.nn.relu <https://www.tensorflow.org/api_docs/python/tf/nn/relu>`_.
            * **load_weights_path** (*str, optional*) -- When given a path, loads weights from file in that path. Default: None
            * **initializer** (:class:`~psychrnn.backend.initializations.WeightInitializer` *or child object, optional*) -- Initializer to use for the network. Default: :class:`~psychrnn.backend.initializations.WeightInitializer` (:data:`params`) if :data:`params` includes :data:`W_rec` or :data:`load_weights_path` as a key, :class:`~psychrnn.backend.initializations.GaussianSpectralRadius` (:data:`params`) otherwise.
            * **W_in_train** (*bool, optional*) -- True if input weights, W_in, are trainable. Default: True
            * **W_rec_train** (*bool, optional*) -- True if recurrent weights, W_rec, are trainable. Default: True
            * **W_out_train** (*bool, optional*) -- True if output weights, W_out, are trainable. Default: True
            * **b_rec_train** (*bool, optional*) -- True if recurrent bias, b_rec, is trainable. Default: True
            * **b_out_train** (*bool, optional*) -- True if output bias, b_out, is trainable. Default: True
            * **init_state_train** (*bool, optional*) -- True if the inital state for the network, init_state, is trainable. Default: True
            * **loss_function** (*str, optional*) -- Which loss function to use. See :class:`psychrnn.backend.loss_functions.LossFunction` for details. Defaults to ``"mean_squared_error"``.



        :Other Dictionary Keys:
            * Any dictionary keys used by the regularizer will be passed onwards to :class:`psychrnn.backend.regularizations.Regularizer`. See :class:`~psychrnn.backend.regularizations.Regularizer` for key names and details.
            * Any dictionary keys used for the loss function will be passed onwards to :class:`psychrnn.backend.loss_functions.LossFunction`. See :class:`~psychrnn.backend.loss_functions.LossFunction` for key names and details.
            * If :data:`initializer` is not set, any dictionary keys used by the initializer will be pased onwards to :class:`WeightInitializer <psychrnn.backend.initializations.WeightInitializer>` if :data:`load_weights_path` is set or :data:`W_rec` is passed in. Otherwise all keys will be passed to :class:`GaussianSpectralRadius <psychrnn.backend.initializations.GaussianSpectralRadius>`
            * If :data:`initializer` is not set and :data:`load_weights_path` is not set, the dictionary entries returned previously by :func:`get_weights` can be passed in to initialize the network. See :class:`WeightInitializer <psychrnn.backend.initializations.WeightInitializer>` for a list and explanation of possible parameters. At a minimum, :data:`W_rec` must be included as a key to make use of this option.
            * If :data:`initializer` is not set and :data:`load_weights_path` is not set, the following keys can be used to set biological connectivity constraints:

                * **input_connectivity** (*ndarray(dtype=float, shape=(* :attr:`N_rec`, :attr:`N_in` *)), optional*) -- Connectivity mask for the input layer. 1 where connected, 0 where unconnected. Default: np.ones((:attr:`N_rec`, :attr:`N_in`)).
                * **rec_connectivity** (*ndarray(dtype=float, shape=(* :attr:`N_rec`, :attr:`N_rec` *)), optional*) -- Connectivity mask for the recurrent layer. 1 where connected, 0 where unconnected. Default: np.ones((:attr:`N_rec`, :attr:`N_rec`)).
                * **output_connectivity** (*ndarray(dtype=float, shape=(* :attr:`N_out`, :attr:`N_rec` *)), optional*) -- Connectivity mask for the output layer. 1 where connected, 0 where unconnected. Default: np.ones((:attr:`N_out`, :attr:`N_rec`)).
                * **autapses** (*bool, optional*) -- If False, self connections are not allowed in N_rec, and diagonal of :data:`rec_connectivity` will be set to 0. Default: True.
                * **dale_ratio** (float, optional) -- Dale's ratio, used to construct Dale_rec and Dale_out. 0 <= dale_ratio <=1 if dale_ratio should be used. ``dale_ratio * N_rec`` recurrent units will be excitatory, the rest will be inhibitory. Default: None
        
        Inferred Parameters:
            * **alpha** (*float*) -- The number of unit time constants per simulation timestep.

    """
    def __init__(self, params):
        self.params = params

        # --------------------------------------------
        # Unique name used to determine variable scope
        # --------------------------------------------
        try:
            self.name = params['name']
        except KeyError:
            print("You must pass a  'name' to RNN")
            raise

        # ----------------------------------
        # Network sizes (tensor dimensions)
        # ----------------------------------
        try:
            N_in = self.N_in = params['N_in']
        except KeyError:
            print("You must pass 'N_in' to RNN")
            raise
        try:
            N_rec = self.N_rec = params['N_rec']
        except KeyError:
            print("You must pass 'N_rec' to RNN")
            raise
        try:
            N_out = self.N_out = params['N_out']
        except KeyError:
            print("You must pass 'N_out' to RNN")
            raise
        try:
            N_steps = self.N_steps = params['N_steps']
        except KeyError:
            print("You must pass 'N_steps' to RNN")
            raise

        # ----------------------------------
        # Physical parameters
        # ----------------------------------
        try:
            self.dt = params['dt']
        except KeyError:
            print("You must pass 'dt' to RNN")
            raise
            
        try:
            self.tau = params['tau']
        except KeyError:
            print("You must pass 'tau' to RNN")
            raise
        try: 
            self.tau = self.tau.astype('float32')
        except AttributeError:
            pass

        try:
            self.N_batch = params['N_batch']
        except KeyError:
            print("You must pass 'N_batch' to RNN")
            raise
            
        self.alpha = (1.0 * self.dt) / self.tau
        self.rec_noise = params.get('rec_noise', 0.0)
        self.transfer_function = params.get('transfer_function', tf.nn.relu)


        # ----------------------------------
        # Load weights path
        # ----------------------------------
        self.load_weights_path = params.get('load_weights_path', None)

        # ------------------------------------------------
        # Define initializer for TensorFlow variables
        # ------------------------------------------------
        if self.load_weights_path is not None:
            self.initializer = WeightInitializer(load_weights_path=self.load_weights_path)
        elif params.get('W_rec', None) is not None:
            self.initializer = params.get('initializer',
                                          WeightInitializer(**params))
        else:
            self.initializer = params.get('initializer',
                                          GaussianSpectralRadius(**params))

        self.dale_ratio = self.initializer.get_dale_ratio()

        # ----------------------------------
        # Trainable features
        # ----------------------------------
        self.W_in_train = params.get('W_in_train', True)
        self.W_rec_train = params.get('W_rec_train', True)
        self.W_out_train = params.get('W_out_train', True)
        self.b_rec_train = params.get('b_rec_train', True)
        self.b_out_train = params.get('b_out_train', True)
        self.init_state_train = params.get('init_state_train', True)

        # --------------------------------------------------
        # TensorFlow input/output placeholder initializations
        # ---------------------------------------------------
        self.x = tf.compat.v1.placeholder("float", [None, N_steps, N_in])
        self.y = tf.compat.v1.placeholder("float", [None, N_steps, N_out])
        self.output_mask = tf.compat.v1.placeholder("float", [None, N_steps, N_out])

        # --------------------------------------------------
        # Initialize variables in proper scope
        # ---------------------------------------------------
        with tf.compat.v1.variable_scope(self.name) as scope:
            # ------------------------------------------------
            # Trainable variables:
            # Initial State, weight matrices and biases
            # ------------------------------------------------
            try:
                self.init_state = tf.compat.v1.get_variable('init_state', [1, N_rec],
                                              initializer=self.initializer.get('init_state'),
                                              trainable=self.init_state_train)
            except ValueError as error:
                raise UserWarning("Try calling model.destruct() or changing params['name'].")


            self.init_state = tf.tile(self.init_state, [self.N_batch, 1])

            # Input weight matrix:
            self.W_in = \
                tf.compat.v1.get_variable('W_in', [N_rec, N_in],
                                initializer=self.initializer.get('W_in'),
                                trainable=self.W_in_train)

            # Recurrent weight matrix:
            self.W_rec = \
                tf.compat.v1.get_variable(
                    'W_rec',
                    [N_rec, N_rec],
                    initializer=self.initializer.get('W_rec'),
                    trainable=self.W_rec_train)

            # Output weight matrix:
            self.W_out = tf.compat.v1.get_variable('W_out', [N_out, N_rec],
                                         initializer=self.initializer.get('W_out'),
                                         trainable=self.W_out_train)

            # Recurrent bias:
            self.b_rec = tf.compat.v1.get_variable('b_rec', [N_rec], initializer=self.initializer.get('b_rec'),
                                         trainable=self.b_rec_train)
            # Output bias:
            self.b_out = tf.compat.v1.get_variable('b_out', [N_out], initializer=self.initializer.get('b_out'),
                                         trainable=self.b_out_train)

            # ------------------------------------------------
            # Non-trainable variables:
            # Overall connectivity and Dale's law matrices
            # ------------------------------------------------

            # Recurrent Dale's law weight matrix:
            self.Dale_rec = tf.compat.v1.get_variable('Dale_rec', [N_rec, N_rec],
                                            initializer=self.initializer.get('Dale_rec'),
                                            trainable=False)

            # Output Dale's law weight matrix:
            self.Dale_out = tf.compat.v1.get_variable('Dale_out', [N_rec, N_rec],
                                            initializer=self.initializer.get('Dale_out'),
                                            trainable=False)

            # Connectivity weight matrices:
            self.input_connectivity = tf.compat.v1.get_variable('input_connectivity', [N_rec, N_in],
                                                      initializer=self.initializer.get('input_connectivity'),
                                                      trainable=False)
            self.rec_connectivity = tf.compat.v1.get_variable('rec_connectivity', [N_rec, N_rec],
                                                    initializer=self.initializer.get('rec_connectivity'),
                                                    trainable=False)
            self.output_connectivity = tf.compat.v1.get_variable('output_connectivity', [N_out, N_rec],
                                                       initializer=self.initializer.get('output_connectivity'),
                                                       trainable=False)

        # --------------------------------------------------
        # Flag to check if variables initialized, model built
        # ---------------------------------------------------
        self.is_initialized = False
        self.is_built = False

    def build(self):
        """ Build the TensorFlow network and start a TensorFlow session.

        """
        # --------------------------------------------------
        # Define the predictions
        # --------------------------------------------------
        self.predictions, self.states = self.forward_pass()

        # --------------------------------------------------
        # Define the loss (based on the predictions)
        # --------------------------------------------------
        self.loss = LossFunction(self.params).set_model_loss(self)

        # --------------------------------------------------
        # Define the regularization
        # --------------------------------------------------
        self.reg = Regularizer(self.params).set_model_regularization(self)

        # --------------------------------------------------
        # Define the total regularized loss
        # --------------------------------------------------
        self.reg_loss = self.loss + self.reg

        # --------------------------------------------------
        # Open a session
        # --------------------------------------------------
        self.sess = tf.compat.v1.Session()

        # --------------------------------------------------
        # Record successful build
        # --------------------------------------------------
        self.is_built = True

        return

    def destruct(self):
        """ Close the TensorFlow session and reset the global default graph.

        """
        # --------------------------------------------------
        # Close the session. Delete the graph.
        # --------------------------------------------------
        if self.is_built:
            self.sess.close()
        tf.compat.v1.reset_default_graph()
        return

    def get_effective_W_rec(self):
        """ Get the recurrent weights used in the network, after masking by connectivity and dale_ratio.

        Returns:
            tf.Tensor(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec` ))

        """
        W_rec = self.W_rec * self.rec_connectivity
        if self.dale_ratio:
            W_rec = tf.matmul(tf.abs(W_rec), self.Dale_rec, name="in_1")
        return W_rec

    def get_effective_W_in(self):
        """ Get the input weights used in the network, after masking by connectivity and dale_ratio.

        Returns:
            tf.Tensor(dtype=float, shape=(:attr:`N_rec`, :attr:`N_in` ))
        """

        W_in = self.W_in * self.input_connectivity
        if self.dale_ratio:
            W_in = tf.abs(W_in)
        return W_in

    def get_effective_W_out(self):
        """ Get the output weights used in the network, after masking by connectivity, and dale_ratio.

        Returns:
            tf.Tensor(dtype=float, shape=(:attr:`N_out`, :attr:`N_rec` ))
        """

        W_out = self.W_out * self.output_connectivity
        if self.dale_ratio:
            W_out = tf.matmul(tf.abs(self.W_out), self.Dale_out, name="in_2")
        return W_out
    
    @abstractmethod
    def forward_pass(self):
        """ Run the RNN on a batch of task inputs. 

        Note:
            This is an abstract function that must be defined in a child class.
        
        Returns: 
            tuple:
            * **predictions** (*ndarray(dtype=float, shape=(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Network output on inputs found in self.x within the tf network.
            * **states** (*ndarray(dtype=float, shape=(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- State variable values over the course of the trials found in self.x within the tf network.

        """
        raise UserWarning("forward_pass must be implemented in child class. See Basic for example.")

    def get_weights(self):
        """ Get weights used in the network. 

        Allows for rebuilding or tweaking different weights to do experiments / analyses.

        Returns:
            dict: Dictionary of rnn weights including the following keys:

            :Dictionary Keys: 
                * **init_state** (*ndarray(dtype=float, shape=(1, :attr:`N_rec` *))*) -- Initial state of the network's recurrent units.
                * **W_in** (*ndarray(dtype=float, shape=(:attr:`N_rec`. :attr:`N_in` *))*) -- Input weights.
                * **W_rec** (*ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec` *))*) -- Recurrent weights.
                * **W_out** (*ndarray(dtype=float, shape=(:attr:`N_out`, :attr:`N_rec` *))*) -- Output weights.
                * **b_rec** (*ndarray(dtype=float, shape=(:attr:`N_rec`, *))*) -- Recurrent bias.
                * **b_out** (*ndarray(dtype=float, shape=(:attr:`N_out`, *))*) -- Output bias.
                * **Dale_rec** (*ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec`*))*) -- Diagonal matrix with ones and negative ones on the diagonal. If :data:`dale_ratio` is not ``None``, indicates whether a recurrent unit is excitatory(1) or inhibitory(-1).
                * **Dale_out** (*ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec`*))*) -- Diagonal matrix with ones and zeroes on the diagonal. If :data:`dale_ratio` is not ``None``, indicates whether a recurrent unit is excitatory(1) or inhibitory(0). Inhibitory neurons do not contribute to the output.
                * **input_connectivity** (*ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_in`*))*) -- Connectivity mask for the input layer. 1 where connected, 0 where unconnected.
                * **rec_connectivity** (*ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec`*))*) -- Connectivity mask for the recurrent layer. 1 where connected, 0 where unconnected.
                * **output_connectivity** (*ndarray(dtype=float, shape=(:attr:`N_out`, :attr:`N_rec`*))*) -- Connectivity mask for the output layer. 1 where connected, 0 where unconnected.
                * **dale_ratio** (*float*) -- Dale's ratio, used to construct Dale_rec and Dale_out. Either ``None`` if dale's law was not applied, or 0 <= dale_ratio <=1 if dale_ratio was applied.

            Note:
                Keys returned may be different / include other keys depending on the implementation of :class:`RNN` used. A different set of keys will be included e.g. if the :class:`~psychrnn.backend.models.lstm.LSTM` implementation is used. The set of keys above is accurate and meaningful for the :class:`~psychrnn.backend.models.basic.Basic` and :class:`~psychrnn.backend.models.basic.BasicScan` implementations.
        """
        if not self.is_built:
            self.build()

        if not self.is_initialized:
            self.sess.run(tf.compat.v1.global_variables_initializer())
            self.is_initialized = True
      
        weights_dict = dict()
        
        for var in tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.GLOBAL_VARIABLES, scope=self.name):
            # avoid saving duplicates
            if var.name.endswith(':0') and var.name.startswith(self.name):
                name = var.name[len(self.name)+1:-2]
                weights_dict.update({name: var.eval(session=self.sess)})
        weights_dict.update({'W_rec': self.get_effective_W_rec().eval(session=self.sess)})
        weights_dict.update({'W_in': self.get_effective_W_in().eval(session=self.sess)})
        weights_dict.update({'W_out': self.get_effective_W_out().eval(session=self.sess)})
        weights_dict['dale_ratio'] = self.dale_ratio
        return weights_dict

    def save(self, save_path):
        """ Save the weights returned by :func:`get_weights` to :data:`save_path`

        Arguments:
            save_path (str): Path for where to save the network weights.

        """

        weights_dict = self.get_weights()

        np.savez(save_path, **weights_dict)

        return

    def train(self, trial_batch_generator, train_params={}):
        """ Train the network.

        Arguments:
            trial_batch_generator (:class:`~psychrnn.tasks.task.Task` object or *Generator[tuple, None, None]*): the task to train on, or the task to train on's batch_generator. If a task is passed in, task.:func:`batch_generator` () will be called to get the generator for the task to train on.
            train_params (dict, optional): Dictionary of training parameters containing the following possible keys:

                :Dictionary Keys: 
                    * **learning_rate** (*float, optional*) -- Sets learning rate if use default optimizer Default: .001
                    * **training_iters** (*int, optional*) -- Number of iterations to train for Default: 50000.
                    * **loss_epoch** (*int, optional*) -- Compute and record loss every 'loss_epoch' epochs. Default: 10.
                    * **verbosity** (*bool, optional*) -- If true, prints information as training progresses. Default: True.
                    * **save_weights_path** (*str, optional*) -- Where to save the model after training. Default: None
                    * **save_training_weights_epoch** (*int, optional*) -- Save training weights every 'save_training_weights_epoch' epochs. Weights only actually saved if :data:`training_weights_path` is set. Default: 100.
                    * **training_weights_path** (*str, optional*) -- What directory to save training weights into as training progresses. Default: None.               
                    * **curriculum** (`~psychrnn.backend.curriculum.Curriculum` *object, optional*) -- Curriculum to train on. If a curriculum object is provided, it overrides the trial_batch_generator argument. Default: None.
                    * **optimizer** (`tf.compat.v1.train.Optimizer <https://www.tensorflow.org/api_docs/python/tf/compat/v1/train/Optimizer>`_ *object, optional*) -- What optimizer to use to compute gradients. Default: `tf.train.AdamOptimizer <https://www.tensorflow.org/api_docs/python/tf/compat/v1/train/AdamOptimizer>`_ (learning_rate=:data:`train_params`['learning_rate']` ).
                    * **clip_grads** (*bool, optional*) -- If true, clip gradients by norm 1. Default: True
                    * **fixed_weights** (*dict, optional*) -- By default all weights are allowed to train unless :data:`fixed_weights` or :data:`W_rec_train`, :data:`W_in_train`, or :data:`W_out_train` are set. Default: None. Dictionary of weights to fix (not allow to train) with the following optional keys:

                        Fixed Weights Dictionary Keys (in case of :class:`~psychrnn.backend.models.basic.Basic` and :class:`~psychrnn.backend.models.basic.BasicScan` implementations)
                            * **W_in** (*ndarray(dtype=bool, shape=(:attr:`N_rec`. :attr:`N_in` *)), optional*) -- True for input weights that should be fixed during training.
                            * **W_rec** (*ndarray(dtype=bool, shape=(:attr:`N_rec`, :attr:`N_rec` *)), optional*) -- True for recurrent weights that should be fixed during training.
                            * **W_out** (*ndarray(dtype=bool, shape=(:attr:`N_out`, :attr:`N_rec` *)), optional*) -- True for output weights that should be fixed during training.

                        :Note:
                            In general, any key in the dictionary output by :func:`get_weights` can have a key in the fixed_weights matrix, however fixed_weights will only meaningfully apply to trainable matrices.

                    * **performance_cutoff** (*float*) -- If :data:`performance_measure` is not ``None``, training stops as soon as performance_measure surpases the performance_cutoff. Default: None.
                    * **performance_measure** (*function*) -- Function to calculate the performance of the network using custom criteria. Default: None.

                        :Arguments:
                            * **trial_batch** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Task stimuli for :attr:`N_batch` trials.
                            * **trial_y** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Target output for the network on :attr:`N_batch` trials given the :data:`trial_batch`.
                            * **output_mask** (*ndarray(dtype=bool, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Output mask for :attr:`N_batch` trials. True when the network should aim to match the target output, False when the target output can be ignored.
                            * **output** (*ndarray(dtype=bool, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Output to compute the accuracy of. ``output`` as returned by :func:`psychrnn.backend.rnn.RNN.test`.
                            * **epoch** (*int*): Current training epoch (e.g. perhaps the performance_measure is calculated differently early on vs late in training)
                            * **losses** (*list of float*): List of losses from the beginning of training until the current epoch.
                            * **verbosity** (*bool*): Passed in from :data:`train_params`.

                        :Returns:
                            *float* 

                            Performance, greater when the performance is better.
        Returns:
            tuple:
            * **losses** (*list of float*) -- List of losses, computed every :data:`loss_epoch` epochs during training.
            * **training_time** (*float*) -- Time spent training.
            * **initialization_time** (*float*) -- Time spent initializing the network and preparing to train.

        """
        if not self.is_built:
            self.build()

        t0 = time()
        # --------------------------------------------------
        # Extract params
        # --------------------------------------------------
        learning_rate = train_params.get('learning_rate', .001)
        training_iters = train_params.get('training_iters', 50000)
        loss_epoch = train_params.get('loss_epoch', 10)
        verbosity = train_params.get('verbosity', True)
        save_weights_path = train_params.get('save_weights_path', None)
        save_training_weights_epoch = train_params.get('save_training_weights_epoch', 100)
        training_weights_path = train_params.get('training_weights_path', None)
        curriculum = train_params.get('curriculum', None)
        optimizer = train_params.get('optimizer',
                                     tf.compat.v1.train.AdamOptimizer(learning_rate=learning_rate))
        clip_grads = train_params.get('clip_grads', True)
        fixed_weights = train_params.get('fixed_weights', None) # array of zeroes and ones. One indicates to pin and not train that weight.
        performance_cutoff = train_params.get('performance_cutoff', None)
        performance_measure = train_params.get('performance_measure', None)

        if (performance_cutoff is not None and performance_measure is None) or (performance_cutoff is None and performance_measure is not None):
                raise UserWarning("training will not be cutoff based on performance. Make sure both performance_measure and performance_cutoff are defined")

        if curriculum is not None:
            trial_batch_generator = curriculum.get_generator_function()

        if not isgenerator(trial_batch_generator):
            trial_batch_generator = trial_batch_generator.batch_generator()

        # --------------------------------------------------
        # Make weights folder if it doesn't already exist.
        # --------------------------------------------------
        if save_weights_path != None:
            if path.dirname(save_weights_path) != "" and not path.exists(path.dirname(save_weights_path)):
                makedirs(path.dirname(save_weights_path))

        # --------------------------------------------------
        # Make train weights folder if it doesn't already exist.
        # --------------------------------------------------
        if training_weights_path != None:
            if path.dirname(training_weights_path) != "" and not path.exists(path.dirname(training_weights_path)):
                makedirs(path.dirname(_weights_path))

        # --------------------------------------------------
        # Compute gradients
        # --------------------------------------------------
        grads = optimizer.compute_gradients(self.reg_loss)

        # --------------------------------------------------
        # Fixed Weights
        # --------------------------------------------------
        if fixed_weights is not None:
            for i in range(len(grads)):
                (grad, var) = grads[i]
                name = var.name[len(self.name)+1:-2]
                if name in fixed_weights.keys():
                    grad = tf.multiply(grad, (1-fixed_weights[name]))
                    grads[i] = (grad, var)


        # --------------------------------------------------
        # Clip gradients
        # --------------------------------------------------
        if clip_grads:
            grads = [(tf.clip_by_norm(grad, 1.0), var)
                     if grad is not None else (grad, var)
                     for grad, var in grads]

        # --------------------------------------------------
        # Call the optimizer and initialize variables
        # --------------------------------------------------
        optimize = optimizer.apply_gradients(grads)
        self.sess.run(tf.compat.v1.global_variables_initializer())
        self.is_initialized = True

        # --------------------------------------------------
        # Record training time for performance benchmarks
        # --------------------------------------------------
        t1 = time()

        # --------------------------------------------------
        # Training loop
        # --------------------------------------------------
        epoch = 1
        batch_size = next(trial_batch_generator)[0].shape[0]
        losses = []
        if performance_cutoff is not None:
            performance = performance_cutoff - 1

        while epoch * batch_size < training_iters and (performance_cutoff is None or performance < performance_cutoff):
            batch_x, batch_y, output_mask, _ = next(trial_batch_generator)
            self.sess.run(optimize, feed_dict={self.x: batch_x, self.y: batch_y, self.output_mask: output_mask})
            # --------------------------------------------------
            # Output batch loss
            # --------------------------------------------------
            if epoch % loss_epoch == 0:
                reg_loss = self.sess.run(self.reg_loss,
                                feed_dict={self.x: batch_x, self.y: batch_y, self.output_mask: output_mask})
                losses.append(reg_loss)
                if verbosity:
                    print("Iter " + str(epoch * batch_size) + ", Minibatch Loss= " + \
                          "{:.6f}".format(reg_loss))

            # --------------------------------------------------
            # Allow for curriculum learning
            # --------------------------------------------------
            if curriculum is not None and epoch % curriculum.metric_epoch == 0:
                trial_batch, trial_y, output_mask, _ = next(trial_batch_generator)
                output, _ = self.test(trial_batch)
                if curriculum.metric_test(trial_batch, trial_y, output_mask, output, epoch, losses, verbosity):
                    if curriculum.stop_training:
                        break
                    trial_batch_generator = curriculum.get_generator_function()

            # --------------------------------------------------
            # Save intermediary weights
            # --------------------------------------------------
            if epoch % save_training_weights_epoch == 0:
                if training_weights_path is not None:
                    self.save(training_weights_path + str(epoch))
                    if verbosity:
                        print("Training weights saved in file: %s" % training_weights_path + str(epoch))
            
            # ---------------------------------------------------
            # Update performance value if necessary
            # ---------------------------------------------------
            if performance_measure is not None:
                trial_batch, trial_y, output_mask, _ = next(trial_batch_generator)
                output, _ = self.test(trial_batch)
                performance = performance_measure(trial_batch, trial_y, output_mask, output, epoch, losses, verbosity)
                if verbosity:
                    print("performance: " + str(performance))
            epoch += 1

        t2 = time()
        if verbosity:
            print("Optimization finished!")

        # --------------------------------------------------
        # Save final weights
        # --------------------------------------------------
        if save_weights_path is not None:
            self.save(save_weights_path)
            if verbosity:
                print("Model saved in file: %s" % save_weights_path)

        # --------------------------------------------------
        # Return losses, training time, initialization time
        # --------------------------------------------------
        return losses, (t2 - t1), (t1 - t0)

    
    def train_curric(self, train_params):
        """Wrapper function for training with curriculum to streamline curriculum learning.

        Arguments: 
            train_params (dict, optional): See :func:`train` for details.

        Returns:
            tuple: See :func:`train` for details.
        """
        # --------------------------------------------------
        # Wrapper function for training with curriculum
        # to streamline curriculum learning
        # --------------------------------------------------

        curriculum = train_params.get('curriculum', None)
        if curriculum is None:
            raise UserWarning("train_curric requires a curriculum. Please pass in a curriculum or use train instead.")
        
        losses, training_time, initialization_time = self.train(curriculum.get_generator_function(), train_params)

        return losses, training_time, initialization_time

    def test(self, trial_batch):
        """ Test the network on a certain task input.

        Arguments:
            trial_batch ((*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Task stimulus to run the network on. Stimulus from :func:`psychrnn.tasks.task.Task.get_trial_batch`, or from next(:func:`psychrnn.tasks.task.Task.batch_generator` ).
        
        Returns:
            tuple:
            * **outputs** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Output time series of the network for each trial in the batch.
            * **states** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- Activity of recurrent units during each trial.
        """
        if not self.is_built:
            self.build()

        if not self.is_initialized:
            self.sess.run(tf.compat.v1.global_variables_initializer())
            self.is_initialized = True

        # --------------------------------------------------
        # Run the forward pass on trial_batch
        # --------------------------------------------------
        outputs, states = self.sess.run([self.predictions, self.states],
                                        feed_dict={self.x: trial_batch})

        return outputs, states


from __future__ import division

from psychrnn.tasks.task import Task
import numpy as np


class MatchToCategory(Task):
    """ Multidirectional decision-making task.

    On each trial the network receives input from units representing different locations on a ring. Each input unit magnitude represents closeness to the angle of input. The network must determine which side of arbitrary category boundaries the input belongs to and respond accordingly.

    Takes :attr:`N_in` channels of noisy input arranged in a ring with gaussian signal around the ring centered at 0 at the stimulus angle.
    :attr:`N_out` channel output arranged as slices of a ring with a one hot encoding towards the correct category output based on the angular location of the gaussian input bump.

    Loosely based on `Freedman, David J., and John A. Assad. "Experience-dependent representation of visual categories in parietal cortex." Nature 443.7107 (2006): 85-88. <https://www.nature.com/articles/nature05078>`_

    Args:
        dt (float): The simulation timestep.
        tau (float): The intrinsic time constant of neural state decay.
        T (float): The trial length.
        N_batch (int): The number of trials per training update.
        N_in (int, optional): The number of network inputs. Defaults to 16.
        N_out (int, optional): The number of network outputs. Defaults to 2.

    """
    
    def __init__(self, dt, tau, T, N_batch, N_in=16, N_out=2):
        super(MatchToCategory,self).__init__(N_in, N_out, dt, tau, T, N_batch)
        
    def _gaussian_input(self,angle, scale = 1):
        """ Calculates angular gaussian pdf with mean at the :data:`angle` for N_in evenly arranged around the circumference of a unit circle.

        Args:
            angle (float): The angle on the input circle at which to center the gaussian.
            scale (float, optional): The scale of the gaussian function. Defaults to 1.

        Returns:
            ndarray(dtype=float, shape=(:attr:`N_in`,)): Normal pdf value at (angle - angle of N_in channel).

        """
        
        g = np.zeros(self.N_in)
        map_g = np.linspace(0,2*np.pi,self.N_in)
        for ii in range(self.N_in):

            # Center N_in around angle, and truncate so -pi <= effective angle <= pi
            effective_angle = map_g[ii] - angle
            if effective_angle > np.pi:
                effective_angle = -2*np.pi + effective_angle
            elif effective_angle < -np.pi:
                effective_angle = 2*np.pi + effective_angle

            x = effective_angle/scale

            g[ii] = np.exp(-x**2/2.0) / (np.sqrt(2*np.pi)*scale)
        return g
            
    
    def generate_trial_params(self, batch, trial):
        """Define parameters for each trial.

        Implements :func:`~psychrnn.tasks.task.Task.generate_trial_params`.

        Args:
            batch (int): The batch number that this trial is part of.
            trial (int): The trial number of the trial within the batch *batch*.

        Returns:
            dict: Dictionary of trial parameters including the following keys:

            :Dictionary Keys: 
                * **angle** (*float*) -- Angle at which to center the gaussian. Randomly selected.
                * **category** (*int*) -- Index of the N_out category channel that contains the :data:`angle`.
                * **onset_time** (*float*) -- Stimulus onset time. Set to 200.
                * **input_dur** (*float*) -- Stimulus duration. Set to 1000.
                * **output_dur** (*float*) -- Output duration. The time given to make a choice. Set to 800.
                * **stim_noise** (*float*) -- Scales the stimlus noise. Set to .1.

        """
        params = dict()

        params['angle'] = 2*np.pi*np.random.rand()
        params['category'] = int(params['angle']/(2*np.pi/self.N_out))
        params['onset_time'] = 200
        params['input_dur'] = 1000.
        params['output_dur'] = 800.
        params['stim_noise'] = .1

        return params

    def trial_function(self, t, params):
        """Compute the trial properties at :data:`time`.

        Implements :func:`~psychrnn.tasks.task.Task.trial_function`.

        Based on the :data:`params` compute the trial stimulus (x_t), correct output (y_t), and mask (mask_t) at :data:`time`.

        Args:
            time (int): The time within the trial (0 <= :data:`time` < :attr:`T`).
            params (dict): The trial params produced by :func:`generate_trial_params`.

        Returns:
            tuple:

            * **x_t** (*ndarray(dtype=float, shape=(*:attr:`N_in` *,))*) -- Trial input at :data:`time` given :data:`params`. For ``params['onset_time'] < time < params['onset_time'] + params['input_dur']`` , gaussian pdf with mean = angle and scale = 1 is added to each input channel based on the channel's angle.
            * **y_t** (*ndarray(dtype=float, shape=(*:attr:`N_out` *,))*) -- Correct trial output at :data:`time` given :data:`params`. 1 in the :data:`params['category']` output channel during the output period defined by :data:`params['output_dur']`, 0 otherwise.
            * **mask_t** (*ndarray(dtype=bool, shape=(*:attr:`N_out` *,))*) -- True if the network should train to match the y_t, False if the network should ignore y_t when training. True during the output period, False otherwise.

        """
        # ----------------------------------
        # Initialize with noise
        # ----------------------------------
        x_t = (1./(2*self.dt))*params['stim_noise']*np.random.randn(self.N_in)
        y_t = np.zeros(self.N_out)
        mask_t = np.zeros(self.N_out)

        # ----------------------------------
        # Retrieve parameters
        # ----------------------------------
        onset_time = params['onset_time']
        input_dur = params['input_dur']
        output_dur = params['output_dur']
        
        angle = params['angle']
        category = params['category']


        if onset_time <= t < onset_time + input_dur:
            x_t += self._gaussian_input(angle)

        if onset_time + input_dur <= t < onset_time + input_dur + output_dur:
            y_t[category] += 1.
            mask_t = np.ones(self.N_out)

        return x_t, y_t, mask_t

    def accuracy_function(self, correct_output, test_output, output_mask):
        """Calculates the accuracy of :data:`test_output`.

        Implements :func:`~psychrnn.tasks.task.Task.accuracy_function`.

        Takes the channel-wise mean of the masked output for each trial. Whichever channel has a greater mean is considered to be the network's "choice".

        Returns:
            float: 0 <= accuracy <= 1. Accuracy is equal to the ratio of trials in which the network made the correct choice as defined above.
        
        """
        chosen = np.argmax(np.mean(test_output*output_mask, axis=1), axis = 1)
        truth = np.argmax(np.mean(correct_output*output_mask, axis = 1), axis = 1)
        return np.mean(np.equal(truth, chosen))


from __future__ import division

import tensorflow as tf

tf.compat.v1.disable_eager_execution()

class Regularizer(object):
    """Regularizer Class

    Class that aggregates all types of regularization used.

    Args:
       params (dict): The regularization parameters containing the following optional keys:

           :Dictionary Keys:
                * **L1_in** (*float, optional*) -- Parameter for weighting the L1 input weights regularization. Default: 0.
                * **L1_rec** (*float, optional*) -- Parameter for weighting the L1 recurrent weights regularization. Default: 0.
                * **L1_out** (*float, optional*) -- Parameter for weighting the L1 output weights regularization. Default: 0.
                * **L2_in** (*float, optional*) -- Parameter for weighting the L2 input weights regularization. Default: 0.
                * **L2_rec** (*float, optional*) -- Parameter for weighting the L2 recurrent weights regularization. Default: 0.
                * **L2_out** (*float, optional*) -- Parameter for weighting the L2 output weights regularization. Default: 0.
                * **L2_firing_rate** (*float, optional*) -- Parameter for weighting the L2 regularization of the relu thresholded states. Default: 0.
                * **custom_regularization** (*function, optional*) -- Custom regularization function. Default: None.

                    Args:
                        * **model** (:class:`~psychrnn.backend.rnn.RNN` *object*) -- Model for which to calculate the regularization.
                        * **params** (*dict*) -- Regularization parameters. All params passed to the :class:`Regularizer` will be passed here.

                    Returns:
                        tf.Tensor(dtype=float)-- The custom regularization to add when calculating the loss.
    """

    def __init__(self, params):
        # ----------------------------------
        # regularization coefficients
        # ----------------------------------
        self.L1_in = params.get('L1_in', 0)
        self.L1_rec = params.get('L1_rec', 0)
        self.L1_out = params.get('L1_out', 0)

        self.L2_in = params.get('L2_in', 0)
        self.L2_rec = params.get('L2_rec', 0)
        self.L2_out = params.get('L2_out', 0)

        self.L2_firing_rate = params.get('L2_firing_rate', 0)

        self.custom_regularization =  params.get('custom_regularization', None)
        self.params = params

    def set_model_regularization(self, model):
        """ Given model, calculate the regularization by adding all regualarization terms (scaled with the parameters to be either zero or nonzero).

        The following regularizations are added: :func:`L1_weight_reg`, :func:`L2_weight_reg`, and :func:`L2_firing_rate_reg`.

        Args:
            model (:class:`~psychrnn.backend.rnn.RNN` object): Model for which to calculate the regularization.

        Returns:
            tf.Tensor(dtype=float): The regularization to add when calculating the loss.
        """
        reg = 0

        # ----------------------------------
        # L1 weight regularization
        # ----------------------------------
        reg += self.L1_weight_reg(model)

        # ----------------------------------
        # L2 weight regularization
        # ----------------------------------
        reg += self.L2_weight_reg(model)

        # ----------------------------------
        # L2 firing rate regularization
        # ----------------------------------
        reg += self.L2_firing_rate_reg(model)


        if self.custom_regularization is not None:
            reg += self.custom_regularization(model, self.params)

        return reg

    def L1_weight_reg(self, model):
        """ L1 regularization

        :math:`regularization = L1\\_in * ||W\\_in||_1 + L1\\_rec * ||W\\_rec||_1 + L1\\_out * ||W\\_out||_1`

        Args:
            model (:class:`~psychrnn.backend.rnn.RNN` object): Model for which to calculate the regularization.

        Returns:
            tf.Tensor(dtype=float): The L1 regularization to add when calculating the loss.
        """

        reg = 0

        reg += self.L1_in * tf.reduce_mean(input_tensor=tf.abs(model.get_effective_W_in()))
        reg += self.L1_rec * tf.reduce_mean(input_tensor=tf.abs(model.get_effective_W_rec()))
        reg += self.L1_out * tf.reduce_mean(input_tensor=tf.abs(model.get_effective_W_out()))

        return reg

    def L2_weight_reg(self, model):
        """ L2 regularization

        :math:`regularization = L2\\_in * ||W\\_in||_2^2 + L2\\_rec * ||W\\_rec||_2^2 + L2\\_out * ||W\\_out||_2^2`

        Args:
            model (:class:`~psychrnn.backend.rnn.RNN` object): Model for which to calculate the regularization.

        Returns:
            tf.Tensor(dtype=float): The L2 regularization to add when calculating the loss.
        """

        reg = 0

        reg += self.L2_in * tf.reduce_mean(input_tensor=tf.square(tf.abs(model.get_effective_W_in())))
        reg += self.L2_rec * tf.reduce_mean(input_tensor=tf.square(tf.abs(model.get_effective_W_rec())))
        reg += self.L2_out * tf.reduce_mean(input_tensor=tf.square(tf.abs(model.get_effective_W_out())))

        return reg

    def L2_firing_rate_reg(self, model):
        """ L2 regularization of the firing rate.

        :math:`regularization = L2\\_firing\\_rate * ||relu(states)||_2^2`

        Args:
            model (:class:`~psychrnn.backend.rnn.RNN` object): Model for which to calculate the regularization.

        Returns:
            tf.Tensor(dtype=float): The L2 firing rate regularization to add when calculating the loss.
        """

        reg = self.L2_firing_rate * tf.reduce_mean(input_tensor=tf.square(model.transfer_function(model.states)))

        return reg



from __future__ import division

from psychrnn.backend.rnn import RNN
import tensorflow as tf

tf.compat.v1.disable_eager_execution()


class LSTM(RNN):
    """ LSTM (Long Short Term Memory) recurrent network model

    LSTM implementation of :class:`psychrnn.backend.rnn.RNN`. Because LSTM is structured differently from the basic RNN, biological constraints such as dale's, autapses, and connectivity are not enabled.

    Args:
       params (dict): See :class:`psychrnn.backend.rnn.RNN` for details.

    """

    def __init__(self, params):
        # ----------------------------------
        # Call RNN constructor
        # ----------------------------------
        super(LSTM, self).__init__(params)

        # ----------------------------------
        # Add new variables for gates
        # TODO better LSTM initialization
        # ----------------------------------
        self.N_concat = self.N_in + self.N_rec

        self.init_hidden_initializer = tf.compat.v1.random_normal_initializer(mean=0, stddev=0.1)
        self.init_cell_initializer = tf.compat.v1.random_normal_initializer(mean=0, stddev=0.1)

        self.W_f_initializer = tf.compat.v1.random_normal_initializer(mean=0, stddev=0.1)
        self.W_i_initializer = tf.compat.v1.random_normal_initializer(mean=0, stddev=0.1)
        self.W_c_initializer = tf.compat.v1.random_normal_initializer(mean=0, stddev=0.1)
        self.W_o_initializer = tf.compat.v1.random_normal_initializer(mean=0, stddev=0.1)

        self.b_f_initializer = tf.compat.v1.constant_initializer(1.0)
        self.b_i_initializer = tf.compat.v1.constant_initializer(1.0)
        self.b_c_initializer = tf.compat.v1.constant_initializer(1.0)
        self.b_o_initializer = tf.compat.v1.constant_initializer(1.0)

        # ----------------------------------
        # TensorFlow initializations
        # ----------------------------------
        with tf.compat.v1.variable_scope(self.name) as scope:
            self.init_hidden = tf.compat.v1.get_variable('init_hidden', [self.N_batch, self.N_rec],
                                               initializer=self.init_hidden_initializer,
                                                trainable=True)
            self.init_cell = tf.compat.v1.get_variable('init_cell', [self.N_batch, self.N_rec],
                                             initializer=self.init_cell_initializer,
                                            trainable=True)

            self.W_f = tf.compat.v1.get_variable('W_f', [self.N_concat, self.N_rec],
                                            initializer=self.W_f_initializer,
                                            trainable=True)
            self.W_i = tf.compat.v1.get_variable('W_i', [self.N_concat, self.N_rec],
                                            initializer=self.W_i_initializer,
                                            trainable=True)
            self.W_c = tf.compat.v1.get_variable('W_c', [self.N_concat, self.N_rec],
                                            initializer=self.W_c_initializer,
                                            trainable=True)
            self.W_o = tf.compat.v1.get_variable('W_o', [self.N_concat, self.N_rec],
                                            initializer=self.W_o_initializer,
                                            trainable=True)

            self.b_f = tf.compat.v1.get_variable('b_f', [self.N_rec], initializer=self.b_f_initializer,
                                         trainable=True)
            self.b_i = tf.compat.v1.get_variable('b_i', [self.N_rec], initializer=self.b_i_initializer,
                                       trainable=True)
            self.b_c = tf.compat.v1.get_variable('b_c', [self.N_rec], initializer=self.b_c_initializer,
                                       trainable=True)
            self.b_o = tf.compat.v1.get_variable('b_o', [self.N_rec], initializer=self.b_o_initializer,
                                       trainable=True)

    def recurrent_timestep(self, rnn_in, hidden, cell):
        """ Recurrent time step.

        Given input and previous state, outputs the next state of the network.

        Arguments:
            rnn_in (*tf.Tensor(dtype=float, shape=(?*, :attr:`N_in` *))*): Input to the rnn at a certain time point.
            hidden (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): Hidden units state of network at previous time point.
            cell (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): Cell state of the network at previous time point.

        Returns:
            tuple:
            * **new_hidden** (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*) -- New hidden unit state of the network.
            * **new_cell** (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*) -- New cell state of the network.
        """

        f = tf.nn.sigmoid(tf.matmul(tf.concat([hidden, rnn_in], 1), self.W_f)
                               + self.b_f)

        i = tf.nn.sigmoid(tf.matmul(tf.concat([hidden, rnn_in], 1), self.W_i)
                               + self.b_i)

        c = tf.nn.tanh(tf.matmul(tf.concat([hidden, rnn_in], 1), self.W_c)
                               + self.b_c)

        o = tf.nn.sigmoid(tf.matmul(tf.concat([hidden, rnn_in], 1), self.W_o)
                          + self.b_o)

        new_cell = f * cell + i * c

        new_hidden = o * tf.nn.sigmoid(new_cell)

        return new_hidden, new_cell

    def output_timestep(self, hidden):
        """Returns the output node activity for a given timestep.

        Arguments:
            hidden (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): Hidden units of network at a given timepoint for each trial in the batch.

        Returns:
            output (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_out` *))*): Output of the network at a given timepoint for each trial in the batch.

        """

        output = tf.matmul(hidden, self.W_out, transpose_b=True) + self.b_out

        return output

    def forward_pass(self):
        """ Run the LSTM on a batch of task inputs.

        Iterates over timesteps, running the :func:`recurrent_timestep` and :func:`output_timestep`

        Implements :func:`psychrnn.backend.rnn.RNN.forward_pass`.

        Returns:
            tuple:
            * **predictions** (*tf.Tensor(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Network output on inputs found in self.x within the tf network.
            * **hidden** (*tf.Tensor(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- Hidden unit values over the course of the trials found in self.x within the tf network.

        """

        rnn_inputs = tf.unstack(self.x, axis=1)
        hidden = self.init_hidden
        cell = self.init_cell
        rnn_outputs = []
        rnn_states = []
        for rnn_input in rnn_inputs:
            hidden, cell = self.recurrent_timestep(rnn_input, hidden, cell)
            output = self.output_timestep(hidden)
            rnn_outputs.append(output)
            rnn_states.append(hidden)

        return tf.transpose(a=rnn_outputs, perm=[1, 0, 2]), tf.transpose(a=rnn_states, perm=[1, 0, 2])



from __future__ import division

import numpy as np
import tensorflow as tf
from warnings import warn

tf.compat.v1.disable_eager_execution()


class WeightInitializer(object):
    """ Base Weight Initialization class.

    Initializes biological constraints and network weights, optionally loading weights from a file or from passed in arrays.

    Keyword Arguments:
        N_in (int): The number of network inputs.
        N_rec (int): The number of recurrent units in the network.
        N_out (int): The number of network outputs.
        load_weights_path (str, optional): Path to load weights from using np.load. Weights saved at that path should be in the form saved out by :func:`psychrnn.backend.rnn.RNN.save` Default: None.
        
        input_connectivity (ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_in`)), optional): Connectivity mask for the input layer. 1 where connected, 0 where unconnected. Default: np.ones((:attr:`N_rec`, :attr:`N_in`)).
        rec_connectivity (ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec`)), optional): Connectivity mask for the recurrent layer. 1 where connected, 0 where unconnected. Default: np.ones((:attr:`N_rec`, :attr:`N_rec`)).
        output_connectivity (ndarray(dtype=float, shape=(:attr:`N_out`, :attr:`N_rec`)), optional): Connectivity mask for the output layer. 1 where connected, 0 where unconnected. Default: np.ones((:attr:`N_out`, :attr:`N_rec`)).
        autapses (bool, optional): If False, self connections are not allowed in N_rec, and diagonal of :data:`rec_connectivity` will be set to 0. Default: True.
        dale_ratio (float, optional): Dale's ratio, used to construct Dale_rec and Dale_out. 0 <= dale_ratio <=1 if dale_ratio should be used. ``dale_ratio * N_rec`` recurrent units will be excitatory, the rest will be inhibitory. Default: None
        
        which_rand_init (str, optional): Which random initialization to use for W_in and W_out. Will also be used for W_rec if :data:`which_rand_W_rec_init` is not passed in. Options: :func:`'const_unif' <const_unif_init>`, :func:`'const_gauss' <const_gauss_init>`, :func:`'glorot_unif' <glorot_unif_init>`, :func:`'glorot_gauss' <glorot_gauss_init>`. Default: :func:`'glorot_gauss' <glorot_gauss_init>`.
        which_rand_W_rec_init (str, optional): Which random initialization to use for W_rec. Options: :func:`'const_unif' <const_unif_init>`, :func:`'const_gauss' <const_gauss_init>`, :func:`'glorot_unif' <glorot_unif_init>`, :func:`'glorot_gauss' <glorot_gauss_init>`. Default: :data:`which_rand_init`.
        init_minval (float, optional): Used by :func:`const_unif_init` as :attr:`minval` if ``'const_unif'`` is passed in for :data:`which_rand_init` or :data:`which_rand_W_rec_init`. Default: -.1.
        init_maxval (float, optional): Used by :func:`const_unif_init` as :attr:`maxval` if ``'const_unif'`` is passed in for :data:`which_rand_init` or :data:`which_rand_W_rec_init`. Default: .1. 

        W_in (ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_in` )), optional): Input weights. Default: Initialized using the function indicated by :data:`which_rand_init`
        W_rec (ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec` )), optional): Recurrent weights. Default: Initialized using the function indicated by :data:`which_rand_W_rec_init`
        W_out (ndarray(dtype=float, shape=(:attr:`N_out`, :attr:`N_rec` )), optional): Output weights. Defualt: Initialized using the function indicated by :data:`which_rand_init`
        b_rec (ndarray(dtype=float, shape=(:attr:`N_rec`, )), optional): Recurrent bias. Default: np.zeros(:attr:`N_rec`)
        b_out (ndarray(dtype=float, shape=(:attr:`N_out`, )), optional): Output bias. Default: np.zeros(:attr:`N_out`)
        Dale_rec (ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec`)), optional): Diagonal matrix with ones and negative ones on the diagonal. If :data:`dale_ratio` is not ``None``, indicates whether a recurrent unit is excitatory(1) or inhibitory(-1). Default: constructed based on :data:`dale_ratio`
        Dale_out (ndarray(dtype=float, shape=(:attr:`N_rec`, :attr:`N_rec`)), optional): Diagonal matrix with ones and zeroes on the diagonal. If :data:`dale_ratio` is not ``None``, indicates whether a recurrent unit is excitatory(1) or inhibitory(0). Inhibitory neurons do not contribute to the output. Default: constructed based on :data:`dale_ratio`
        init_state (ndarray(dtype=float, shape=(1, :attr:`N_rec` )), optional): Initial state of the network's recurrent units. Default: .1 + .01 * np.random.randn(:data:`N_rec` ).

    Attributes:
        initializations (dict): Dictionary containing entries for :data:`input_connectivity`, :data:`rec_connectivity`, :data:`output_connectivity`, :data:`dale_ratio`, :data:`Dale_rec`, :data:`Dale_out`, :data:`W_in`, :data:`W_rec`, :data:`W_out`, :data:`b_rec`, :data:`b_out`, and :data:`init_state`.
    
    """


    def __init__(self, **kwargs):

        # ----------------------------------
        # Required parameters
        # ----------------------------------
        self.load_weights_path = kwargs.get('load_weights_path', None)
        N_in = self.N_in = kwargs.get('N_in')
        N_rec = self.N_rec = kwargs.get('N_rec')
        N_out = self.N_out = kwargs.get('N_out')
        self.autapses = kwargs.get('autapses', True)

        self.initializations = dict()

        if self.load_weights_path is not None:
            # ----------------------------------
            # Load saved weights
            # ----------------------------------
            self.initializations = dict(np.load(self.load_weights_path, allow_pickle = True))
            if type(self.initializations['dale_ratio']) == np.ndarray:
                self.initializations['dale_ratio'] = self.initializations['dale_ratio'].item()

        else:
            if kwargs.get('W_rec', None) is None and type(self).__name__=='WeightInitializer':
                warn("This network may not train since the eigenvalues of W_rec are not regulated in any way.")

            # ----------------------------------
            # Optional Parameters
            # ----------------------------------
            self.rand_init = kwargs.get('which_rand_init', 'glorot_gauss')
            self.rand_W_rec_init = self.get_rand_init_func(kwargs.get('which_rand_W_rec_init', self.rand_init))
            self.rand_init = self.get_rand_init_func(self.rand_init)
            if self.rand_init == self.const_unif_init or self.rand_W_rec_init == self.const_unif_init:
                self.init_minval = kwargs.get('init_minval', -.1)
                self.init_maxval = kwargs.get('init_maxval', .1)

            # ----------------------------------
            # Biological Constraints
            # ----------------------------------

            # Connectivity constraints
            self.initializations['input_connectivity'] = kwargs.get('input_connectivity',np.ones([N_rec, N_in]))
            assert(self.initializations['input_connectivity'].shape == (N_rec, N_in))
            self.initializations['rec_connectivity'] = kwargs.get('rec_connectivity',np.ones([N_rec, N_rec]))
            assert(self.initializations['rec_connectivity'].shape == (N_rec, N_rec))
            self.initializations['output_connectivity'] = kwargs.get('output_connectivity', np.ones([N_out, N_rec]))
            assert(self.initializations['output_connectivity'].shape == (N_out, N_rec))
            
            # Autapses constraint
            if not self.autapses:
                self.initializations['rec_connectivity'][np.eye(N_rec) == 1] = 0

            # Dale's constraint
            self.initializations['dale_ratio'] = dale_ratio = kwargs.get('dale_ratio', None)
            if type(self.initializations['dale_ratio']) == np.ndarray:
                self.initializations['dale_ratio'] = dale_ratio = self.initializations['dale_ratio'].item()
            if dale_ratio is not None and (dale_ratio <0 or dale_ratio > 1):
                print("Need 0 <= dale_ratio <= 1. dale_ratio was: " + str(dale_ratio))
                raise
            dale_vec = np.ones(N_rec)
            if dale_ratio is not None:
                dale_vec[int(dale_ratio * N_rec):] = -1
                dale_rec = np.diag(dale_vec)
                dale_vec[int(dale_ratio * N_rec):] = 0
                dale_out = np.diag(dale_vec)
            else:
                dale_rec = np.diag(dale_vec)
                dale_out = np.diag(dale_vec)
            self.initializations['Dale_rec'] = kwargs.get('Dale_rec', dale_rec)
            assert(self.initializations['Dale_rec'].shape == (N_rec, N_rec))
            self.initializations['Dale_out'] = kwargs.get('Dale_out', dale_rec)
            assert(self.initializations['Dale_out'].shape == (N_rec, N_rec))

            # ----------------------------------
            # Default initializations / optional loading from params
            # ----------------------------------

            self.initializations['W_in'] = kwargs.get('W_in', self.rand_init(self.initializations['input_connectivity']))
            assert(self.initializations['W_in'].shape == (N_rec, N_in))
            self.initializations['W_out'] = kwargs.get('W_out', self.rand_init(self.initializations['output_connectivity']))
            assert(self.initializations['W_out'].shape == (N_out, N_rec))
            self.initializations['W_rec'] = kwargs.get('W_rec', self.rand_W_rec_init(self.initializations['rec_connectivity']))
            assert(self.initializations['W_rec'].shape == (N_rec, N_rec))

            self.initializations['b_rec'] = kwargs.get('b_rec',np.zeros(N_rec))
            assert(self.initializations['b_rec'].shape == (N_rec,))
            self.initializations['b_out'] = kwargs.get('b_out',np.zeros(N_out))
            assert(self.initializations['b_out'].shape == (N_out,))

            self.initializations['init_state'] = kwargs.get('init_state', .1 + .01 * np.random.randn(N_rec))
            assert(self.initializations['init_state'].size == N_rec)

        return

    def get_rand_init_func(self, which_rand_init):
        """Maps initialization function names (strings) to generating functions.

        Arguments:
            which_rand_init (str): Maps to ``[which_rand_init]_init``. Options are :func:`'const_unif' <const_unif_init>`, :func:`'const_gauss' <const_gauss_init>`, :func:`'glorot_unif' <glorot_unif_init>`, :func:`'glorot_gauss' <glorot_gauss_init>`.

        Returns:
            function: ``self.[which_rand_init]_init``

        """
        mapping = {
            'const_unif': self.const_unif_init,
            'const_gauss': self.const_gauss_init,
            'glorot_unif': self.glorot_unif_init,
            'glorot_gauss': self.glorot_gauss_init}
        return mapping[which_rand_init]

    def const_gauss_init(self, connectivity):
        """ Initialize ndarray of shape :data:`connectivity` with values from a normal distribution.

        Arguments:
            connectivity (ndarray): 1 where connected, 0 where unconnected.

        Returns:
            ndarray(dtype=float, shape=connectivity.shape)

        """
        return np.random.randn(connectivity.shape[0], connectivity.shape[1])

    def const_unif_init(self, connectivity):
        """ Initialize ndarray of shape :data:`connectivity` with values uniform distribution with minimum :data:`init_minval` and maximum :data:`init_maxval` as set in :class:`WeightInitializer`.

        Arguments:
            connectivity (ndarray): 1 where connected, 0 where unconnected.

        Returns:
            ndarray(dtype=float, shape=connectivity.shape)

        """
        minval = self.init_minval
        maxval = self.init_maxval
        return (maxval-minval) * np.random.rand(connectivity.shape[0], connectivity.shape[1]) + minval

    def glorot_unif_init(self, connectivity):
        """ Initialize ndarray of shape :data:`connectivity` with values from a glorot uniform distribution.

        Draws samples from a uniform distribution within [-limit, limit] where `limit`
        is `sqrt(6 / (fan_in + fan_out))` where `fan_in` is the number of input units and `fan_out` is the number of output units. Respects the :data:`connectivity` matrix.

        Arguments:
            connectivity (ndarray): 1 where connected, 0 where unconnected.

        Returns:
            ndarray(dtype=float, shape=connectivity.shape)

        """


        init = np.zeros(connectivity.shape)
        fan_in = np.sum(connectivity, axis = 1)
        init += np.tile(fan_in, (connectivity.shape[1],1)).T
        fan_out = np.sum(connectivity, axis = 0)
        init += np.tile(fan_out, (connectivity.shape[0],1))
        return np.random.uniform(-np.sqrt(6/init), np.sqrt(6/init))

    def glorot_gauss_init(self, connectivity):
        """ Initialize ndarray of shape :data:`connectivity` with values from a glorot normal distribution.

        Draws samples from a normal distribution centered on 0 with `stddev
        = sqrt(2 / (fan_in + fan_out))` where `fan_in` is the number of input units and `fan_out` is the number of output units. Respects the :data:`connectivity` matrix.

        Arguments:
            connectivity (ndarray): 1 where connected, 0 where unconnected.

        Returns:
            ndarray(dtype=float, shape=connectivity.shape)

        """

        init = np.zeros(connectivity.shape)
        fan_in = np.sum(connectivity, axis = 1)
        init += np.tile(fan_in, (connectivity.shape[1],1)).T
        fan_out = np.sum(connectivity, axis = 0)
        init += np.tile(fan_out, (connectivity.shape[0],1))
        return np.random.normal(0, np.sqrt(2/init))

    def get_dale_ratio(self):
        """ Returns the dale_ratio.

        :math:`0 \\leq dale\\_ratio \\leq 1` if dale_ratio should be used, dale_ratio = None otherwise. ``dale_ratio * N_rec`` recurrent units will be excitatory, the rest will be inhibitory.

        Returns:
            float: Dale ratio, None if no dale ratio is set.
        """

        return self.initializations['dale_ratio']

    def get(self, tensor_name):
        """ Get :data:`tensor_name` from :attr:`initializations` as a Tensor.

        Arguments:
            tensor_name (str): The name of the tensor to get. See :attr:`initializations` for options.

        Returns:
            Tensor object

        """

        return tf.compat.v1.constant_initializer(self.initializations[tensor_name])

    def save(self, save_path):
        """ Save :attr:`initializations` to :data:`save_path`.

        Arguments:
            save_path (str): File path for saving the initializations. The .npz extension will be appended if not already provided.
        """

        np.savez(save_path, **self.initializations)
        return

    def balance_dale_ratio(self):
        """ If dale_ratio is not None, balances :attr:`initializations['W_rec'] <initializations>` 's excitatory and inhibitory weights so the network will train. 
        """
        dale_ratio = self.get_dale_ratio()
        if dale_ratio is not None:
            dale_vec = np.ones(self.N_rec)
            dale_vec[int(dale_ratio * self.N_rec):] = dale_ratio/(1-dale_ratio)
            dale_rec = np.diag(dale_vec) / np.linalg.norm(np.matmul(self.initializations['rec_connectivity'], np.diag(dale_vec)), axis=1)[:,np.newaxis]
            self.initializations['W_rec'] = np.matmul(self.initializations['W_rec'], dale_rec)
        return






class GaussianSpectralRadius(WeightInitializer):
    """Generate random gaussian weights with specified spectral radius.

    If Dale is set, balances the random gaussian weights between excitatory and inhibitory using :func:`balance_dale_ratio` before applying the specified spectral radius.

    Keyword Args:
       spec_rad (float, optional): The spectral radius to initialize W_rec with. Default: 1.1.

    Other Keyword Args:
        See :class:`~psychrnn.backend.initializations.WeightInitializer` for details.
    """

    def __init__(self, **kwargs):

        super(GaussianSpectralRadius, self).__init__(**kwargs)

        self.spec_rad = kwargs.get('spec_rad', 1.1)

        self.initializations['W_rec']  = np.random.randn(self.N_rec, self.N_rec)

        # balance weights for dale ratio training to proceed normally
        self.balance_dale_ratio()

        self.initializations['W_rec'] = self.spec_rad * self.initializations['W_rec'] / np.max(np.abs(np.linalg.eig(self.initializations['W_rec'])[0]))

        return



class AlphaIdentity(WeightInitializer):
    '''Generate recurrent weights :math:`w(i,i) = alpha`, :math:`w(i,j) = 0` where :math:`i \\neq j`.

    If Dale is set, balances the alpha excitatory and inhibitory weights using :func:`~psychrnn.backend.initializations.WeightInitializer.balance_dale_ratio`, so w(i,i) will not be exactly equal to alpha.

    Keyword Args:
       alpha (float): The value of alpha to set w(i,i) to in W_rec.

    Other Keyword Args:
        See :class:`WeightInitializer` for details.
    '''

    def __init__(self, **kwargs):

        super(AlphaIdentity, self).__init__(**kwargs)

        self.alpha = kwargs.get('alpha')

        self.initializations['W_rec'] = np.eye(self.N_rec) * self.alpha
        
        # balance weights for dale ratio training to proceed normally
        self.balance_dale_ratio()

        return



from __future__ import division

import tensorflow as tf
tf.compat.v1.disable_eager_execution()


class LossFunction(object):
    """ Set the loss function for the :class:`~psychrnn.backend.rnn.RNN` model.

    Arguments:
        params(dict): Dictionary of parameters including the following keys:

            :Dictionary Keys:
                * **loss_function** (*str*) -- String indicating what loss function to use. If :data:`params["loss_function"]` is not `mean_squared_error` or `binary_cross_entropy`, :data:`params[params["loss_function"]]` defines the custom loss function. Default: "mean_squared_error".
                * **params["loss_function"]** (*function, optional*) -- Defines the custom loss function. Must have the same signature as :func:`mean_squared_error` and :func:`binary_cross_entropy`.

    """

    def __init__(self, params):
        self.type = params.get("loss_function", "mean_squared_error")
        if self.type != "mean_squared_error" and self.type != "binary_cross_entropy":
            self.custom_loss_function = params.get(self.type, None)
            if self.custom_loss_function is None:
                raise UserWarning("Loss type is '" + self.type + "' but '" + self.type + "' is not an entry in params. Did you mean 'mean_square_error' or 'binary_cross_entropy'? If not, you must pass a function in to params as '" + self.type + "'." )

    def set_model_loss(self, model):
        """ Returns the model loss, calculated as indicated by :attr:`type` (inferred from :data:`params["loss_function"]`.

        ``'mean_squared_error'`` indicates :func:`mean_squared_error`, ``'binary_cross_entropy'`` indicates :func:`binary_cross_entropy`.
        If :attr:`type` is not one of the above options, :attr:`custom_loss_function` is used. The custom loss function would have been passed in to :data:`params` as :data:`params[type]`.

        Args:
            model (:class:`~psychrnn.backend.rnn.RNN` object): Model for which to calculate the regularization.

        Returns:
            tf.Tensor(dtype=float): Model loss.

        """

        loss = 0

        if self.type == "mean_squared_error":
            loss = self.mean_squared_error(model.predictions, model.y, model.output_mask)

        elif self.type == "binary_cross_entropy":
            loss = self.binary_cross_entropy(model.predictions, model.y, model.output_mask)

        else:
            loss = self.custom_loss_function(model.predictions, model.y, model.output_mask)

        return loss

    def mean_squared_error(self, predictions, y, output_mask):
        """ Mean squared error.

        ``loss = mean(square(output_mask * (predictions - y)))``

        Args:
            predictions (*tf.Tensor(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Network output.
            y (*tf.Tensor(dtype=float, shape =(*?, :attr:`N_steps`, :attr:`N_out` *))*): Target output.
            output_mask (*tf.Tensor(dtype=float, shape =(*?, :attr:`N_steps`, :attr:`N_out` *))*): Output mask for :attr:`N_batch` trials. True when the network should aim to match the target output, False when the target output can be ignored.

        Returns:
            tf.Tensor(dtype=float): Mean squared error.

        """

        return tf.reduce_mean(input_tensor=tf.square(output_mask * (predictions - y)))

    def binary_cross_entropy(self, predictions, y, output_mask):
        """ Binary cross-entropy.

        Binary label values are assumed to be 0 and 1. 

        ``loss = mean(output_mask * -(y * log(predictions) + (1-y)* log(1-predictions)))``

        Args:
            predictions (*tf.Tensor(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Network output.
            y (*tf.Tensor(dtype=float, shape =(*?, :attr:`N_steps`, :attr:`N_out` *))*): Target output.
            output_mask (*tf.Tensor(dtype=float, shape =(*?, :attr:`N_steps`, :attr:`N_out` *))*): Output mask for :attr:`N_batch` trials. True when the network should aim to match the target output, False when the target output can be ignored.

        Returns:
            tf.Tensor(dtype=float): Binary cross-entropy.

        """

        epsilon = 1e-07 # default epsilon used in TensorFlow
        predictions = tf.clip_by_value(predictions, epsilon, 1. - epsilon)

        return tf.reduce_mean( input_tensor=output_mask *
                               -(y * tf.math.log(predictions + epsilon) + (1 - y) * tf.math.log(1 - predictions + epsilon)))






from __future__ import division

from psychrnn.tasks.task import Task
import numpy as np

class DelayedDiscrimination(Task):
    """Delayed discrimination task. 

    Following a fore period, the network receives an input, followed by a delay. After the delay the network receives a second input. The second input channel receives noisy input that is inversely ordered compared to the input received by the first input channel. The network must respond by activating the output node that corresponds to the input channel with the greater input as the first stimulus.

    Takes two channels of noisy input (:attr:`N_in` = 2).
    Two channel output (:attr:`N_out` = 2) with a one hot encoding (high value is 1, low value is .2).

    Loosely based on `Romo, R., Brody, C. D., Hernández, A., & Lemus, L. (1999). Neuronal correlates of 
    parametric working memory in the prefrontal cortex. Nature, 399(6735), 470. <https://www.nature.com/articles/20939>`_

    Args:
        dt (float): The simulation timestep.
        tau (float): The intrinsic time constant of neural state decay.
        T (float): The trial length.
        N_batch (int): The number of trials per training update.
        onset_time (float, optional): Stimulus onset time in terms of trial length :data:`T`.
        stim_duration_1 (float, optional): Stimulus 1 duration in terms of trial length :data:`T`.
        delay_duration (float, optional): Delay duration in terms of trial length :data:`T`.
        stim_duration_2 (float, optional): Stimulus 2 duration in terms of trial length :data:`T`.
        decision_duration (float, optional): Decision duration  in terms of trial length :data:`T`.
    """

    def __init__(self, dt, tau, T, N_batch, onset_time = None, stim_duration_1 = None, delay_duration = None, stim_duration_2 = None, decision_duration = None):
        super(DelayedDiscrimination,self).__init__(2, 2, dt, tau, T, N_batch)
        
        self.onset_time = onset_time
        self.stim_duration_1 = stim_duration_1
        self.delay_duration = delay_duration
        self.stim_duration_2 = stim_duration_2
        self.decision_duration = decision_duration 

        self.frequency_pairs = [(18, 10), (22, 14), (26, 18), (30, 22), (34, 26)] # frequency pairs to select from
        self.decision_options = ['>', '<'] # decision options to select from

        self.lo = 0.2 # Low value for one hot encoding
        self.hi = 1.0 # High value for one hot encoding

    def _scale_p(self, f):
        """Scale frequency to be between .4 and 1.2."""
        return 0.4 + 0.8 * (f - 10) / (34 - 10)

    def _scale_n(self, f):
        """ Scale frequency to be between .4 and 1.2, invert frequency ordering."""
        return 0.4 + 0.8 * (34 - f) / (34 - 10)

    def generate_trial_params(self, batch, trial):
        """Define parameters for each trial.

        Implements :func:`~psychrnn.tasks.task.Task.generate_trial_params`.

        Args:
            batch (int): The batch number that this trial is part of.
            trial (int): The trial number of the trial within the batch *batch*.

        Returns:
            dict: Dictionary of trial parameters including the following keys:

            :Dictionary Keys: 
                * **stimulus_1** (*float*) -- Start time for stimulus one. :data:`onset_time`. 
                * **delay** (*float*) -- Start time for the delay. :data:`onset_time` + :data:`stimulus_duration_1`. 
                * **stimulus_2** (*float*) -- Start time in for stimulus one. :data:`onset_time` + :data:`stimulus_duration_1` + :data:`delay_duration`. 
                * **decision** (*float*) -- Start time in for decision period. :data:`onset_time` + :data:`stimulus_duration_1` + :data:`delay_duration` + :data:`stimulus_duration_2`. 
                * **end** (*float*) -- End of decision period. :data:`onset_time` + :data:`stimulus_duration_1` + :data:`delay_duration` + :data:`stimulus_duration_2` + :data:`decision_duration`. 
                * **stim_noise** (*float*) -- Scales the stimlus noise. Set to .1.
                * **f1** (*int*) -- Frequency of first stimulus. 
                * **f2** (*int*) -- Frequency of second stimulus.
                * **choice** (*str*) -- Indicates whether :data:`f1` is '>' or '<' :data:`f2`.
        """

        params = dict()

        if self.onset_time is None:
            onset_time = np.random.uniform(0, 1) * self.T / 8.0
        else:
            onset_time = self.onset_time
        
        if self.stim_duration_1 is None:
            stim_duration_1 = np.random.uniform(0, 1) * self.T / 4.0
        else:
            stim_duration_1 = self.stim_duration_1
        
        if self.delay_duration is None:
            delay_duration = np.random.uniform(0, 1) * self.T / 4.0
        else:
            delay_duration = self.delay_duration
        
        if self.stim_duration_2 is None:
            stim_duration_2 = np.random.uniform(0, 1) * self.T / 4.0
        else:
            stim_duration_2 = self.stim_duration_2
       
        if self.decision_duration is None:
            decision_duration = np.random.uniform(0, 1) * self.T / 8.0
        else:
            decision_duration = self.decision_duration
        
        params['stimulus_1'] = onset_time
        params['delay'] = onset_time + stim_duration_1
        params['stimulus_2'] = onset_time + stim_duration_1 + delay_duration
        params['decision'] = onset_time + stim_duration_1 + delay_duration + stim_duration_2
        params['end'] = onset_time + stim_duration_1 + delay_duration + stim_duration_2 + decision_duration

        params['stim_noise'] = 0.1

        fpair = self.frequency_pairs[np.random.choice(len(self.frequency_pairs))]
        gt_lt = np.random.choice(self.decision_options)
        if gt_lt == '>':
            f1, f2 = fpair
            choice = 0
        else:
            f2, f1 = fpair
            choice = 1
        params['f1'] = f1
        params['f2'] = f2
        params['choice'] = choice

        return params

    def trial_function(self, t, params):
        """Compute the trial properties at :data:`time`.

        Implements :func:`~psychrnn.tasks.task.Task.trial_function`.

        Based on the :data:`params` compute the trial stimulus (x_t), correct output (y_t), and mask (mask_t) at :data:`time`.

        Args:
            time (int): The time within the trial (0 <= :data:`time` < :attr:`T`).
            params (dict): The trial params produced by :func:`generate_trial_params`.

        Returns:
            tuple:

            * **x_t** (*ndarray(dtype=float, shape=(*:attr:`N_in` *,))*) -- Trial input at :data:`time` given :data:`params`. First channel contains :data:`f1` during the first stimulus period, and :data:`f2` during the second stimulus period, scaled to be between .4 and 1.2. Second channel contains the frequencies but reverse scaled -- high frequencies correspond to low values and vice versa. Both channels have baseline noise.
            * **y_t** (*ndarray(dtype=float, shape=(*:attr:`N_out` *,))*) -- Correct trial output at :data:`time` given :data:`params`. The correct output is encoded using one-hot encoding during the decision period.
            * **mask_t** (*ndarray(dtype=bool, shape=(*:attr:`N_out` *,))*) -- True if the network should train to match the y_t, False if the network should ignore y_t when training. The mask is True for during the decision period and False otherwise.
        
        """
        # ----------------------------------
        # Initialize with noise
        # ----------------------------------
        x_t = np.sqrt(2*.01*np.sqrt(10)*np.sqrt(self.dt)*params['stim_noise']*params['stim_noise'])*np.random.randn(self.N_in)
        y_t = np.zeros(self.N_out)
        mask_t = np.zeros(self.N_out)

        # ----------------------------------
        # Retrieve parameters
        # ----------------------------------
        stimulus_1 = params['stimulus_1']
        delay = params['delay']
        stimulus_2 = params['stimulus_2']
        decision = params['decision']
        end = params['end']

        f1 = params['f1']
        f2 = params['f2']
        choice = params['choice']

        # ----------------------------------
        # Compute values
        # ----------------------------------

        if stimulus_1 <= t < delay:
            x_t[0] += self._scale_p(f1)
            x_t[1] += self._scale_n(f1)

        if stimulus_2 <= t < decision:
            x_t[0] += self._scale_p(f2)
            x_t[1] += self._scale_n(f2)

        if decision <= t < end:
            y_t[choice] = self.hi
            y_t[1-choice] = self.lo
            mask_t = np.ones(self.N_out)

        return x_t, y_t, mask_t

    def accuracy_function(self, correct_output, test_output, output_mask):
        """Calculates the accuracy of :data:`test_output`.

        Implements :func:`~psychrnn.tasks.task.Task.accuracy_function`.

        Takes the channel-wise mean of the masked output for each trial. Whichever channel has a greater mean is considered to be the network's "choice".

        Returns:
            float: 0 <= accuracy <= 1. Accuracy is equal to the ratio of trials in which the network made the correct choice as defined above.
        
        """
        
        chosen = np.argmax(np.mean(test_output*output_mask, axis=1), axis = 1)
        truth = np.argmax(np.mean(correct_output*output_mask, axis = 1), axis = 1)
        return np.mean(np.equal(truth, chosen))



import setuptools
from distutils.core import setup

with open("psychrnn/_version.py", "r") as f:
    exec(f.read())

with open("README.md", "r") as f:
    long_desc = f.read()

setup(

    name='PsychRNN',
    version=__version__,
    packages=['psychrnn', 'psychrnn.tasks', 'psychrnn.backend', 'psychrnn.backend.models'],
    license='MIT',
    long_description=long_desc,
    long_description_content_type='text/markdown',

    author="Daniel Ehrlich, Jasmine Stone, Alex Atanasov, David Brandfonbrener",
    author_email="psychrnn@gmail.com",
    description="Easy-to-use package for the modeling and analysis of neural network dynamics, directed towards cognitive neuroscientists.",
    keywords="neuroscience, modeling, analysis, neural networks",
    url="https://github.com/murraylab/PsychRNN",
    project_urls={
    'Documentation': 'https://psychrnn.readthedocs.io/',
    'Mailing List': 'https://www.freelists.org/list/psychrnn',
    },

    install_requires=['python_version>="2.7",!="3.0*",!="3.1*",!="3.2*",!="3.3*"', 'tensorflow']

)


from __future__ import division

from psychrnn.backend.rnn import RNN
import tensorflow as tf
tf.compat.v1.disable_eager_execution()


class Basic(RNN):
    """ The basic continuous time recurrent neural network model.

    Basic implementation of :class:`psychrnn.backend.rnn.RNN` with a simple RNN, enabling biological constraints.

    Args:
       params (dict): See :class:`psychrnn.backend.rnn.RNN` for details.

    """

    def recurrent_timestep(self, rnn_in, state):
        """ Recurrent time step.

        Given input and previous state, outputs the next state of the network.

        Arguments:
            rnn_in (*tf.Tensor(dtype=float, shape=(?*, :attr:`N_in` *))*): Input to the rnn at a certain time point.
            state (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): State of network at previous time point.

        Returns:
            new_state (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): New state of the network.

        """

        new_state = ((1-self.alpha) * state) \
                    + self.alpha * (
                        tf.matmul(
                            self.transfer_function(state),
                            self.get_effective_W_rec(),
                            transpose_b=True, name="1")
                        + tf.matmul(
                            rnn_in,
                            self.get_effective_W_in(),
                            transpose_b=True, name="2")
                        + self.b_rec)\
                    + tf.sqrt(2.0 * self.alpha * self.rec_noise * self.rec_noise)\
                      * tf.random.normal(tf.shape(input=state), mean=0.0, stddev=1.0)

        return new_state

    def output_timestep(self, state):
        """Returns the output node activity for a given timestep.

        Arguments:
            state (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): State of network at a given timepoint for each trial in the batch.

        Returns:
            output (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_out` *))*): Output of the network at a given timepoint for each trial in the batch.

        """
     
        output = tf.matmul(self.transfer_function(state),
                                self.get_effective_W_out(), transpose_b=True, name="3") \
                    + self.b_out
     
        return output

    def forward_pass(self):

        """ Run the RNN on a batch of task inputs.

        Iterates over timesteps, running the :func:`recurrent_timestep` and :func:`output_timestep`

        Implements :func:`psychrnn.backend.rnn.RNN.forward_pass`.

        Returns:
            tuple:
            * **predictions** (*tf.Tensor(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Network output on inputs found in self.x within the tf network.
            * **states** (*tf.Tensor(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- State variable values over the course of the trials found in self.x within the tf network.

        """

        rnn_inputs = tf.unstack(self.x, axis=1)
        state = self.init_state
        rnn_outputs = []
        rnn_states = []
        for rnn_input in rnn_inputs:
            state = self.recurrent_timestep(rnn_input, state)
            output = self.output_timestep(state)
            rnn_outputs.append(output)
            rnn_states.append(state)
        return tf.transpose(a=rnn_outputs, perm=[1, 0, 2]), tf.transpose(a=rnn_states, perm=[1, 0, 2])


class BasicScan(Basic):
    """ The basic continuous time recurrent neural network model implemented with `tf.scan <https://www.tensorflow.org/api_docs/python/tf/scan>`_ .

    Produces the same results as :class:`Basic`, with possible differences in execution time.

    Args:
       params (dict): See :class:`psychrnn.backend.rnn.RNN` for details.

    """

    def recurrent_timestep(self, state, rnn_in):
        """ Wrapper function for :func:`psychrnn.backend.models.basic.Basic.recurrent_timestep`. 

        Arguments:
            state (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): State of network at previous time point.
            rnn_in (*tf.Tensor(dtype=float, shape=(?*, :attr:`N_in` *))*): Input to the rnn at a certain time point.
 
        Returns:
            new_state (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): New state of the network.

        """

        return super(BasicScan, self).recurrent_timestep(rnn_in, state)

    def output_timestep(self, dummy, state):
        """ Wrapper function for :func:`psychrnn.backend.models.basic.Basic.output_timestep`.

        Includes additional dummy argument to facilitate `tf.scan <https://www.tensorflow.org/api_docs/python/tf/scan>`_.

        Arguments:
            dummy: Dummy variable provided by tf.scan. Not actually used by the function.
            state (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*): State of network at a given timepoint for each trial in the batch.

        Returns:
            output (*tf.Tensor(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_out` *))*): Output of the network at a given timepoint for each trial in the batch.

        """
        return super(BasicScan, self).output_timestep(state)

    def forward_pass(self):
        """ Run the RNN on a batch of task inputs.

        Iterates over timesteps, running the :func:`recurrent_timestep` and :func:`output_timestep`

        Implements :func:`psychrnn.backend.rnn.RNN.forward_pass`.

        Returns:
            tuple:
            * **predictions** (*tf.Tensor(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Network output on inputs found in self.x within the tf network.
            * **states** (*tf.Tensor(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- State variable values over the course of the trials found in self.x within the tf network.

        """

        state = self.init_state
        rnn_states = \
            tf.scan(
                self.recurrent_timestep,
                tf.transpose(a=self.x, perm=[1, 0, 2]),
                initializer=state,
                parallel_iterations=1)
        rnn_outputs = \
            tf.scan(
                self.output_timestep,
                rnn_states,
                initializer=tf.zeros([self.N_batch, self.N_out]),
                parallel_iterations=1)
        return tf.transpose(a=rnn_outputs, perm=[1, 0, 2]), tf.transpose(a=rnn_states, perm=[1, 0, 2])


from __future__ import division

import numpy as np

from abc import ABCMeta, abstractmethod

# abstract class python 2 & 3 compatible
ABC = ABCMeta('ABC', (object,), {})

def relu(x):
    """NumPy implementation of `tf.nn.relu <https://www.tensorflow.org/api_docs/python/tf/nn/relu>`_

    Arguments:
        x (ndarray): array for which relu is computed.

    Returns:
        ndarray: np.maximum(x,0)
    """
    return np.maximum(x, 0)

def sigmoid(x):
    """NumPy implementation of `tf.nn.sigmoid <https://www.tensorflow.org/api_docs/python/tf/math/sigmoid>`_

    Arguments:
        x (ndarray): array for which sigmoid is computed.

    Returns:
        ndarray: 1/(1 + np.exp(-x))
    """
    return 1/(1 + np.exp(-x))

class Simulator(ABC):
    """The simulator class.

    Note:
        The base Simulator class is not itself a functioning Simulator. 
        run_trials and rnn_step must be implemented to define a functioning Simulator

    Args:
       rnn_model (:class:`psychrnn.backend.rnn.RNN` object, optional): Uses the :class:`psychrnn.backend.rnn.RNN` object to set :attr:`alpha` and :attr:`rec_noise`. Also used to initialize weights if :data:`weights` and :data:`weights_path` are not passed in. Default: None.
       weights_path (str, optional): Where to load weights from. Take precedence over rnn_model weights. Default: :data:`rnn_model.get_weights() <rnn_model>`. np.load(:data:`weights_path`) should return something of the form :data:`weights`.
       transfer_function (function, optonal): Function that takes an ndarray as input and outputs an ndarray of the same shape with the transfer / activation function applied. NumPy implementation of a TensorFlow transfer function. Default: :func:`relu`.
       weights (dict, optional): Takes precedence over both weights_path and rnn_model. Default: np.load(:data:`weights_path`). Dictionary containing the following keys:

        :Dictionary Keys:
            * **init_state** (*ndarray(dtype=float, shape=(1,* :attr:`N_rec` *))*) -- Initial state of the network's recurrent units.
            * **W_in** (*ndarray(dtype=float, shape=(*:attr:`N_rec`. :attr:`N_in` *))*) -- Input weights.
            * **W_rec** (*ndarray(dtype=float, shape=(*:attr:`N_rec`, :attr:`N_rec` *))*) -- Recurrent weights.
            * **W_out** (*ndarray(dtype=float, shape=(*:attr:`N_out`, :attr:`N_rec` *))*) -- Output weights.
            * **b_rec** (*ndarray(dtype=float, shape=(*:attr:`N_rec`, *))*) -- Recurrent bias.
            * **b_out** (*ndarray(dtype=float, shape=(*:attr:`N_out`, *))*) -- Output bias.

       params (dict, optional):
        :Dictionary Keys:
            * **rec_noise** (*float, optional*) -- Amount of recurrent noise to add to the network. Default: 0
            * **alpha** (*float, optional*) -- The number of unit time constants per simulation timestep. Defaut: (1.0* dt) / tau
            * **dt** (*float, optional*) -- The simulation timestep. Used to calculate alpha if alpha is not passed in. Required if alpha is not in params and rnn_model is None.
            * **tau** (*float*) -- The intrinsic time constant of neural state decay. Used to calculate alpha if alpha is not passed in. Required if alpha is not in params and rnn_model is None.

    """
    def __init__(self,  rnn_model = None, params = None, weights_path=None, weights=None, transfer_function = relu):

        # ----------------------------------
        # Extract params
        # ----------------------------------
        self.transfer_function = transfer_function

        if rnn_model is not None:
            self.alpha = rnn_model.alpha
            self.rec_noise = rnn_model.rec_noise
            if rnn_model.transfer_function.__name__ != self.transfer_function.__name__:
                raise UserWarning("The rnn_model transfer function is " + str(rnn_model.transfer_function) + " and the current transfer function is " + str(self.transfer_function) + ". You should make sure these functions do the same thing -- their names do not match.")
            if params is not None:
                raise UserWarning("params was passed in but will not be used. rnn_model takes precedence.")
        else:
            self.rec_noise = params.get('rec_noise', 0)
            if params.get('alpha') is not None:
                self.alpha = params['alpha']
            else:
                dt = params['dt']
                tau = params['tau']
                self.alpha = params.get('alpha',(1.0* dt) / tau)


        # ----------------------------------
        # Initialize weights
        # ----------------------------------
        self.weights = weights
        if weights is not None:
            if weights_path is not None or rnn_model is not None:
                raise UserWarning("Weights and either rnn_model or weights_path were passed in. Weights from rnn_model and weights_path will be ignored.")
        elif weights_path is not None:
            if rnn_model is not None:
                raise UserWarning("rnn_model and weights_path were both passed in. Weights from rnn_model will be ignored.")
            self.weights = np.load(weights_path)
        elif rnn_model is not None:
            self.weights = rnn_model.get_weights()
        else:
            raise UserWarning("Either weights, rnn_model, or weights_path must be passed in.")

        self.W_in = self.weights['W_in']
        self.W_rec = self.weights['W_rec']
        self.W_out = self.weights['W_out']

        self.b_rec = self.weights['b_rec']
        self.b_out = self.weights['b_out']

        self.init_state = self.weights['init_state']

    # ----------------------------------------------
    # t_connectivity allows for ablation experiments
    # ----------------------------------------------
    @abstractmethod
    def rnn_step(self, state, rnn_in, t_connectivity):
        """Given input and previous state, outputs the next state and output of the network.

        Note:
            This is an abstract function that must be defined in a child class.

        Arguments:
            state (ndarray(dtype=float, shape=(:attr:`N_batch` , :attr:`N_rec`))): State of network at previous time point.
            rnn_in (ndarray(dtype=float, shape=(:attr:`N_batch` , :attr:`N_in`))): State of network at previous time point.
            t_connectivity (ndarray(dtype=float, shape=(:attr:`N_rec` , :attr:`N_rec`))): Matrix for ablating / perturbing W_rec.

        Returns:
            tuple:
            * **new_output** (*ndarray(dtype=float, shape=(*:attr:`N_batch`, :attr:`N_out` *))*) -- Output of the network at a given timepoint for each trial in the batch.
            * **new_state** (*ndarray(dtype=float, shape=(*:attr:`N_batch`, :attr:`N_rec` *))*) -- New state of the network for each trial in the batch.

        """
        pass

    @abstractmethod
    def run_trials(self, trial_input, t_connectivity=None):
        """Test the network on a certain task input, optionally including ablation terms.

        A NumPy implementation of :func:`~psychrnn.backend.rnn.RNN.test` with additional options for ablation.

        N_batch here is flexible and will be inferred from trial_input.

        Arguments:
            trial_batch ((*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Task stimulus to run the network on. Stimulus from :func:`psychrnn.tasks.task.Task.get_trial_batch`, or from next(:func:`psychrnn.tasks.task.Task.batch_generator` ). If you want the network to run autonomously, without input, set input to an array of zeroes, N_steps will still indicate how long to run the network.
            t_connectivity ((*ndarray(dtype=float, shape =(*:attr:`N_steps`, :attr:`N_rec`, :attr:`N_rec` *))*): Matrix for ablating / perturbing W_rec. Passed step by step to rnn_step.

        Returns:
            tuple:
            * **outputs** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Output time series of the network for each trial in the batch.
            * **states** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- Activity of recurrent units during each trial.
        """
        pass


class BasicSimulator(Simulator):
    """:class:`Simulator` implementation for :class:`psychrnn.backend.models.basic.Basic` and for :class:`psychrnn.backend.models.basic.BasicScan`.

    See :class:`Simulator` for arguments.
    """

    def rnn_step(self, state, rnn_in, t_connectivity):
        """Given input and previous state, outputs the next state and output of the network as a NumPy implementation of :class:`psychrnn.backend.models.basic.Basic.recurrent_timestep` and of :class:`psychrnn.backend.models.basic.Basic.output_timestep`.

        Additionally takes in :data:`t_connectivity`. If :data:`t_connectivity` is all ones, :func:`rnn_step`'s output will match that of :class:`psychrnn.backend.models.basic.Basic.recurrent_timestep` and :class:`psychrnn.backend.models.basic.Basic.output_timestep`. Otherwise :data:`W_rec` is multiplied by :data:`t_connectivity` elementwise, ablating / perturbing the recurrent connectivity.

        Arguments:
            state (ndarray(dtype=float, shape=(:attr:`N_batch` , :attr:`N_rec`))): State of network at previous time point.
            rnn_in (ndarray(dtype=float, shape=(:attr:`N_batch` , :attr:`N_in`))): State of network at previous time point.
            t_connectivity (ndarray(dtype=float, shape=(:attr:`N_rec` , :attr:`N_rec`))): Matrix for ablating / perturbing W_rec.

        Returns:
            tuple:
            * **new_output** (*ndarray(dtype=float, shape=(*:attr:`N_batch`, :attr:`N_out` *))*) -- Output of the network at a given timepoint for each trial in the batch.
            * **new_state** (*ndarray(dtype=float, shape=(*:attr:`N_batch`, :attr:`N_rec` *))*) -- New state of the network for each trial in the batch.

        """

        new_state = ((1-self.alpha) * state) \
                    + self.alpha * (
                        np.matmul(
                            self.transfer_function(state),
                            np.transpose(self.W_rec * t_connectivity))
                        + np.matmul(rnn_in, np.transpose(self.W_in))
                        + self.b_rec)\
                    + np.sqrt(2.0 * self.alpha * self.rec_noise * self.rec_noise) * \
                      np.random.normal(loc=0.0, scale=1.0, size=state.shape)

        new_output = np.matmul(
                        self.transfer_function(new_state),
                        np.transpose(self.W_out)) + self.b_out

        return new_output, new_state

    def run_trials(self, trial_input, t_connectivity=None):
        """Test the network on a certain task input, optionally including ablation terms.

        A NumPy implementation of :func:`~psychrnn.backend.rnn.RNN.test` with additional options for ablation.

        N_batch here is flexible and will be inferred from trial_input.

        Repeatedly calls :func:`rnn_step` to build output and states over the entire timecourse of the :data:`trial_batch`

        Arguments:
            trial_batch ((*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Task stimulus to run the network on. Stimulus from :func:`psychrnn.tasks.task.Task.get_trial_batch`, or from next(:func:`psychrnn.tasks.task.Task.batch_generator` ).
                   To run the network autonomously without input, set input to an array of zeroes. N_steps will still indicate for how many steps to run the network.
            t_connectivity ((*ndarray(dtype=float, shape =(*:attr:`N_steps`, :attr:`N_rec`, :attr:`N_rec` *))*): Matrix for ablating / perturbing W_rec. Passed step by step to :func:`rnn_step`.

        Returns:
            tuple:
            * **outputs** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Output time series of the network for each trial in the batch.
            * **states** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- Activity of recurrent units during each trial.
        """

        batch_size = trial_input.shape[0]
        rnn_inputs = np.squeeze(np.split(trial_input, trial_input.shape[1], axis=1))
        state = np.expand_dims(self.init_state[0, :], 0)
        state = np.repeat(state, batch_size, 0)
        rnn_outputs = []
        rnn_states = []
        for i, rnn_input in enumerate(rnn_inputs):

            if t_connectivity is not None:
                output, state = self.rnn_step(state, rnn_input, t_connectivity[i])
            else:
                output, state = self.rnn_step(state, rnn_input, np.ones_like(self.W_rec))

            rnn_outputs.append(output)
            rnn_states.append(state)

        return np.swapaxes(np.array(rnn_outputs), 0, 1), np.swapaxes(np.array(rnn_states), 0, 1)


class LSTMSimulator(Simulator):
    """:class:`Simulator` implementation for :class:`psychrnn.backend.models.lstm.LSTM` and for :class:`psychrnn.backend.models.lstm.LSTM`.

    See :class:`Simulator` for arguments.

    The contents of weights / np.load(weights_path) must now include the following additional keys:

    :Dictionary Keys:
        * **init_hidden** (*ndarray(dtype=float, shape=(*:attr:`N_batch` , :attr:`N_rec` *))*) -- Initial state of the cell state.
        * **init_hidden** (*ndarray(dtype=float, shape=(*:attr:`N_batch` , :attr:`N_rec` *))*) -- Initial state of the hidden state.
        * **W_f** (*ndarray(dtype=float, shape=(*:attr:`N_rec` + :attr:`N_in`, :attr:`N_rec` *))*) -- f term weights
        * **W_i** (*ndarray(dtype=float, shape=(*:attr:`N_rec` + :attr:`N_in`, :attr:`N_rec` *))*) -- i term weights
        * **W_c** (*ndarray(dtype=float, shape=(*:attr:`N_rec` + :attr:`N_in`, :attr:`N_rec` *))*) -- c term weights
        * **W_o** (*ndarray(dtype=float, shape=(*:attr:`N_rec` + :attr:`N_in`, :attr:`N_rec` *))*) -- o term weights
        * **b_f** (*ndarray(dtype=float, shape=(*:attr:`N_rec`, *))*) -- f term bias.
        * **b_i** (*ndarray(dtype=float, shape=(*:attr:`N_rec`, *))*) -- i term bias.
        * **b_c** (*ndarray(dtype=float, shape=(*:attr:`N_rec`, *))*) -- c term bias.
        * **b_o** (*ndarray(dtype=float, shape=(*:attr:`N_rec`, *))*) -- o term bias.
    """

    def __init__(self,  rnn_model = None, params = None, weights_path=None, weights=None):

        super(LSTMSimulator, self).__init__(rnn_model=rnn_model, params=params, weights_path=weights_path, weights=weights)

        self.init_hidden = self.weights['init_hidden']
        self.init_cell = self.weights['init_cell']

        self.W_f = self.weights['W_f']
        self.W_i = self.weights['W_i']
        self.W_c = self.weights['W_c']
        self.W_o = self.weights['W_o']

        self.b_f = self.weights['b_f']
        self.b_i = self.weights['b_i']
        self.b_c = self.weights['b_c']
        self.b_o = self.weights['b_o']

    def rnn_step(self, hidden, cell, rnn_in):
        """Given input and previous state, outputs the next state and output of the network as a NumPy implementation of :class:`psychrnn.backend.models.lstm.LSTM.recurrent_timestep` and of :class:`psychrnn.backend.models.lstm.LSTM.output_timestep`.

        Arguments:
            hidden (ndarray(dtype=float, shape=(:attr:`N_batch` , :attr:`N_rec` ))): Hidden units state of network at previous time point.
            cell (ndarray(dtype=float, shape=(:attr:`N_batch` , :attr:`N_rec` ))): Cell state of the network at previous time point.
            rnn_in (ndarray(dtype=float, shape=(:attr:`N_batch` , :attr:`N_in`))): State of network at previous time point.

        Returns:
            tuple:
            * **new_output** (*ndarray(dtype=float, shape=(*:attr:`N_batch`, :attr:`N_out` *))*) -- Output of the network at a given timepoint for each trial in the batch.
            * **new_hidden** (*ndarray(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*) -- New hidden unit state of the network.
            * **new_cell** (*ndarray(dtype=float, shape=(* :attr:`N_batch` , :attr:`N_rec` *))*) -- New cell state of the network.

        """

        f = sigmoid(np.matmul(np.concatenate([hidden, rnn_in], 1), self.W_f)
                               + self.b_f)

        i = sigmoid(np.matmul(np.concatenate([hidden, rnn_in], 1), self.W_i)
                               + self.b_i)

        c = np.tanh(np.matmul(np.concatenate([hidden, rnn_in], 1), self.W_c)
                               + self.b_c)

        o = sigmoid(np.matmul(np.concatenate([hidden, rnn_in], 1), self.W_o)
                          + self.b_o)

        new_cell = f * cell + i * c

        new_hidden = o * sigmoid(new_cell)

        new_output = np.matmul(new_hidden, np.transpose(self.W_out)) + self.b_out

        return new_output, new_hidden, new_cell

    def run_trials(self, trial_input):
        """Test the network on a certain task input, optionally including ablation terms.

        A NumPy implementation of :func:`~psychrnn.backend.rnn.RNN.test` with additional options for ablation.

        N_batch here is flexible and will be inferred from trial_input.

        Repeatedly calls :func:`rnn_step` to build output and states over the entire timecourse of the :data:`trial_batch`

        Arguments:
            trial_batch ((*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*): Task stimulus to run the network on. Stimulus from :func:`psychrnn.tasks.task.Task.get_trial_batch`, or from next(:func:`psychrnn.tasks.task.Task.batch_generator` ).
                   To run the network autonomously without input, set input to an array of zeroes. N_steps will still indicate for how many steps to run the network.

        Returns:
            tuple:
            * **outputs** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_out` *))*) -- Output time series of the network for each trial in the batch.
            * **states** (*ndarray(dtype=float, shape =(*:attr:`N_batch`, :attr:`N_steps`, :attr:`N_rec` *))*) -- Activity of recurrent units during each trial.
        """

        batch_size = trial_input.shape[0]
        rnn_inputs = np.squeeze(np.split(trial_input, trial_input.shape[1], axis=1))
        cell = np.expand_dims(self.init_cell[0, :], 0)
        cell = np.repeat(cell, batch_size, 0)
        hidden = np.expand_dims(self.init_hidden[0, :], 0)
        hidden = np.repeat(hidden, batch_size, 0)
        rnn_outputs = []
        rnn_states = []
        for i, rnn_input in enumerate(rnn_inputs):

            output, hidden, cell = self.rnn_step(hidden, cell, rnn_input)

            rnn_outputs.append(output)
            rnn_states.append(hidden)

        return np.swapaxes(np.array(rnn_outputs), 0, 1), np.swapaxes(np.array(rnn_states), 0, 1)
